import asyncio
import json
import logging
from contextlib import suppress
from typing import Any, Dict, List, Optional, Tuple, Union

from graphql import ExecutionResult

from ..graphql_request import GraphQLRequest
from .common.adapters.connection import AdapterConnection
from .common.base import SubscriptionTransportBase
from .exceptions import (
    TransportConnectionFailed,
    TransportProtocolError,
    TransportQueryError,
    TransportServerError,
)

log = logging.getLogger("gql.transport.websockets")


class WebsocketsProtocolTransportBase(SubscriptionTransportBase):
    """:ref:`Async Transport <async_transports>` used to execute GraphQL queries on
    remote servers with websocket connection.

    This transport uses asyncio and the provided websockets adapter library
    in order to send requests on a websocket connection.
    """

    # This transport supports two subprotocols and will autodetect the
    # subprotocol supported on the server
    APOLLO_SUBPROTOCOL = "graphql-ws"
    GRAPHQLWS_SUBPROTOCOL = "graphql-transport-ws"

    def __init__(
        self,
        *,
        adapter: AdapterConnection,
        init_payload: Optional[Dict[str, Any]] = None,
        connect_timeout: Optional[Union[int, float]] = 10,
        close_timeout: Optional[Union[int, float]] = 10,
        ack_timeout: Optional[Union[int, float]] = 10,
        keep_alive_timeout: Optional[Union[int, float]] = None,
        ping_interval: Optional[Union[int, float]] = None,
        pong_timeout: Optional[Union[int, float]] = None,
        answer_pings: bool = True,
        subprotocols: Optional[List[str]] = None,
    ) -> None:
        """Initialize the transport with the given parameters.

        :param adapter: The connection dependency adapter
        :param init_payload: Dict of the payload sent in the connection_init message.
        :param connect_timeout: Timeout in seconds for the establishment
            of the websocket connection. If None is provided this will wait forever.
        :param close_timeout: Timeout in seconds for the close. If None is provided
            this will wait forever.
        :param ack_timeout: Timeout in seconds to wait for the connection_ack message
            from the server. If None is provided this will wait forever.
        :param keep_alive_timeout: Optional Timeout in seconds to receive
            a sign of liveness from the server.
        :param ping_interval: Delay in seconds between pings sent by the client to
            the backend for the graphql-ws protocol. None (by default) means that
            we don't send pings. Note: there are also pings sent by the underlying
            websockets protocol. See the
            :ref:`keepalive documentation <websockets_transport_keepalives>`
            for more information about this.
        :param pong_timeout: Delay in seconds to receive a pong from the backend
            after we sent a ping (only for the graphql-ws protocol).
            By default equal to half of the ping_interval.
        :param answer_pings: Whether the client answers the pings from the backend
            (for the graphql-ws protocol).
            By default: True
        :param subprotocols: list of subprotocols sent to the
            backend in the 'subprotocols' http header.
            By default: both apollo and graphql-ws subprotocols.
        """

        if subprotocols is None:
            subprotocols = [
                self.APOLLO_SUBPROTOCOL,
                self.GRAPHQLWS_SUBPROTOCOL,
            ]

        self.adapter.subprotocols = subprotocols

        # Initialize the generic SubscriptionTransportBase parent class
        super().__init__(
            adapter=self.adapter,
            connect_timeout=connect_timeout,
            close_timeout=close_timeout,
            keep_alive_timeout=keep_alive_timeout,
        )

        if init_payload is None:
            init_payload = {}

        self.init_payload: Dict[str, Any] = init_payload
        self.ack_timeout: Optional[Union[int, float]] = ack_timeout

        self.payloads: Dict[str, Any] = {}
        """payloads is a dict which will contain the payloads received
        for example with the graphql-ws protocol: 'ping', 'pong', 'connection_ack'"""

        self.ping_interval: Optional[Union[int, float]] = ping_interval
        self.pong_timeout: Optional[Union[int, float]]
        self.answer_pings: bool = answer_pings

        if ping_interval is not None:
            if pong_timeout is None:
                self.pong_timeout = ping_interval / 2
            else:
                self.pong_timeout = pong_timeout

        self.send_ping_task: Optional[asyncio.Future] = None

        self.ping_received: asyncio.Event = asyncio.Event()
        """ping_received is an asyncio Event which will fire  each time
        a ping is received with the graphql-ws protocol"""

        self.pong_received: asyncio.Event = asyncio.Event()
        """pong_received is an asyncio Event which will fire  each time
        a pong is received with the graphql-ws protocol"""

    async def _wait_ack(self) -> None:
        """Wait for the connection_ack message. Keep alive messages are ignored"""

        while True:
            init_answer = await self._receive()

            answer_type, answer_id, execution_result = self._parse_answer(init_answer)

            if answer_type == "connection_ack":
                return

            if answer_type != "ka":
                raise TransportProtocolError(
                    "Websocket server did not return a connection ack"
                )

    async def _send_init_message_and_wait_ack(self) -> None:
        """Send init message to the provided websocket and wait for the connection ACK.

        If the answer is not a connection_ack message, we will return an Exception.
        """

        init_message = json.dumps(
            {"type": "connection_init", "payload": self.init_payload}
        )

        await self._send(init_message)

        # Wait for the connection_ack message or raise a TimeoutError
        await asyncio.wait_for(self._wait_ack(), self.ack_timeout)

    async def _initialize(self):
        await self._send_init_message_and_wait_ack()

    async def send_ping(self, payload: Optional[Any] = None) -> None:
        """Send a ping message for the graphql-ws protocol"""

        ping_message = {"type": "ping"}

        if payload is not None:
            ping_message["payload"] = payload

        await self._send(json.dumps(ping_message))

    async def send_pong(self, payload: Optional[Any] = None) -> None:
        """Send a pong message for the graphql-ws protocol"""

        pong_message = {"type": "pong"}

        if payload is not None:
            pong_message["payload"] = payload

        await self._send(json.dumps(pong_message))

    async def _send_stop_message(self, query_id: int) -> None:
        """Send stop message to the provided websocket connection and query_id.

        The server should afterwards return a 'complete' message.
        """

        stop_message = json.dumps({"id": str(query_id), "type": "stop"})

        await self._send(stop_message)

    async def _send_complete_message(self, query_id: int) -> None:
        """Send a complete message for the provided query_id.

        This is only for the graphql-ws protocol.
        """

        complete_message = json.dumps({"id": str(query_id), "type": "complete"})

        await self._send(complete_message)

    async def _stop_listener(self, query_id: int) -> None:
        """Stop the listener corresponding to the query_id depending on the
        detected backend protocol.

        For apollo: send a "stop" message
                    (a "complete" message will be sent from the backend)

        For graphql-ws: send a "complete" message and simulate the reception
                        of a "complete" message from the backend
        """
        log.debug(f"stop listener {query_id}")

        if self.subprotocol == self.GRAPHQLWS_SUBPROTOCOL:
            await self._send_complete_message(query_id)
            await self.listeners[query_id].put(("complete", None))
        else:
            await self._send_stop_message(query_id)

    async def _send_connection_terminate_message(self) -> None:
        """Send a connection_terminate message to the provided websocket connection.

        This message indicates that the connection will disconnect.
        """

        connection_terminate_message = json.dumps({"type": "connection_terminate"})

        await self._send(connection_terminate_message)

    async def _send_query(
        self,
        request: GraphQLRequest,
    ) -> int:
        """Send a query to the provided websocket connection.

        We use an incremented id to reference the query.

        Returns the used id for this query.
        """

        query_id = self.next_query_id
        self.next_query_id += 1

        payload: Dict[str, Any] = request.payload

        query_type = "start"

        if self.subprotocol == self.GRAPHQLWS_SUBPROTOCOL:
            query_type = "subscribe"

        query_str = json.dumps(
            {"id": str(query_id), "type": query_type, "payload": payload}
        )

        await self._send(query_str)

        return query_id

    async def _connection_terminate(self):
        if self.subprotocol == self.APOLLO_SUBPROTOCOL:
            await self._send_connection_terminate_message()

    def _parse_answer_graphqlws(
        self, json_answer: Dict[str, Any]
    ) -> Tuple[str, Optional[int], Optional[ExecutionResult]]:
        """Parse the answer received from the server if the server supports the
        graphql-ws protocol.

        Returns a list consisting of:
            - the answer_type (between:
              'connection_ack', 'ping', 'pong', 'data', 'error', 'complete')
            - the answer id (Integer) if received or None
            - an execution Result if the answer_type is 'data' or None

        Differences with the apollo websockets protocol (superclass):
            - the "data" message is now called "next"
            - the "stop" message is now called "complete"
            - there is no connection_terminate or connection_error messages
            - instead of a unidirectional keep-alive (ka) message from server to client,
              there is now the possibility to send bidirectional ping/pong messages
            - connection_ack has an optional payload
            - the 'error' answer type returns a list of errors instead of a single error
        """

        answer_type: str = ""
        answer_id: Optional[int] = None
        execution_result: Optional[ExecutionResult] = None

        try:
            answer_type = str(json_answer.get("type"))

            if answer_type in ["next", "error", "complete"]:
                answer_id = int(str(json_answer.get("id")))

                if answer_type == "next" or answer_type == "error":

                    payload = json_answer.get("payload")

                    if answer_type == "next":

                        if not isinstance(payload, dict):
                            raise ValueError("payload is not a dict")

                        if "errors" not in payload and "data" not in payload:
                            raise ValueError(
                                "payload does not contain 'data' or 'errors' fields"
                            )

                        execution_result = ExecutionResult(
                            errors=payload.get("errors"),
                            data=payload.get("data"),
                            extensions=payload.get("extensions"),
                        )

                        # Saving answer_type as 'data' to be understood with superclass
                        answer_type = "data"

                    elif answer_type == "error":

                        if not isinstance(payload, list):
                            raise ValueError("payload is not a list")

                        raise TransportQueryError(
                            str(payload[0]), query_id=answer_id, errors=payload
                        )

            elif answer_type in ["ping", "pong", "connection_ack"]:
                self.payloads[answer_type] = json_answer.get("payload", None)

            else:
                raise ValueError

            if self.check_keep_alive_task is not None:
                self._next_keep_alive_message.set()

        except ValueError as e:
            raise TransportProtocolError(
                f"Server did not return a GraphQL result: {json_answer}"
            ) from e

        return answer_type, answer_id, execution_result

    def _parse_answer_apollo(
        self, json_answer: Dict[str, Any]
    ) -> Tuple[str, Optional[int], Optional[ExecutionResult]]:
        """Parse the answer received from the server if the server supports the
        apollo websockets protocol.

        Returns a list consisting of:
            - the answer_type (between:
              'connection_ack', 'ka', 'connection_error', 'data', 'error', 'complete')
            - the answer id (Integer) if received or None
            - an execution Result if the answer_type is 'data' or None
        """

        answer_type: str = ""
        answer_id: Optional[int] = None
        execution_result: Optional[ExecutionResult] = None

        try:
            answer_type = str(json_answer.get("type"))

            if answer_type in ["data", "error", "complete"]:
                answer_id = int(str(json_answer.get("id")))

                if answer_type == "data" or answer_type == "error":

                    payload = json_answer.get("payload")

                    if not isinstance(payload, dict):
                        raise ValueError("payload is not a dict")

                    if answer_type == "data":

                        if "errors" not in payload and "data" not in payload:
                            raise ValueError(
                                "payload does not contain 'data' or 'errors' fields"
                            )

                        execution_result = ExecutionResult(
                            errors=payload.get("errors"),
                            data=payload.get("data"),
                            extensions=payload.get("extensions"),
                        )

                    elif answer_type == "error":

                        raise TransportQueryError(
                            str(payload), query_id=answer_id, errors=[payload]
                        )

            elif answer_type == "ka":
                # Keep-alive message
                if self.check_keep_alive_task is not None:
                    self._next_keep_alive_message.set()
            elif answer_type == "connection_ack":
                pass
            elif answer_type == "connection_error":
                error_payload = json_answer.get("payload")
                raise TransportServerError(f"Server error: '{repr(error_payload)}'")
            else:
                raise ValueError

        except ValueError as e:
            raise TransportProtocolError(
                f"Server did not return a GraphQL result: {json_answer}"
            ) from e

        return answer_type, answer_id, execution_result

    def _parse_answer(
        self, answer: str
    ) -> Tuple[str, Optional[int], Optional[ExecutionResult]]:
        """Parse the answer received from the server depending on
        the detected subprotocol.
        """
        try:
            json_answer = json.loads(answer)
        except ValueError:
            raise TransportProtocolError(
                f"Server did not return a GraphQL result: {answer}"
            )

        if self.subprotocol == self.GRAPHQLWS_SUBPROTOCOL:
            return self._parse_answer_graphqlws(json_answer)

        return self._parse_answer_apollo(json_answer)

    async def _send_ping_coro(self) -> None:
        """Coroutine to periodically send a ping from the client to the backend.

        Only used for the graphql-ws protocol.

        Send a ping every ping_interval seconds.
        Close the connection if a pong is not received within pong_timeout seconds.
        """

        assert self.ping_interval is not None

        try:
            while True:
                await asyncio.sleep(self.ping_interval)

                await self.send_ping()

                await asyncio.wait_for(self.pong_received.wait(), self.pong_timeout)

                # Reset for the next iteration
                self.pong_received.clear()

        except asyncio.TimeoutError:
            # No pong received in the appriopriate time, close with error
            # If the timeout happens during a close already in progress, do nothing
            if self.close_task is None:
                await self._fail(
                    TransportServerError(
                        f"No pong received after {self.pong_timeout!r} seconds"
                    ),
                    clean_close=False,
                )

    async def _handle_answer(
        self,
        answer_type: str,
        answer_id: Optional[int],
        execution_result: Optional[ExecutionResult],
    ) -> None:

        # Put the answer in the queue
        await super()._handle_answer(answer_type, answer_id, execution_result)

        # Answer pong to ping for graphql-ws protocol
        if answer_type == "ping":
            self.ping_received.set()
            if self.answer_pings:
                await self.send_pong()

        elif answer_type == "pong":
            self.pong_received.set()

    async def _after_connect(self):

        # Find the backend subprotocol returned in the response headers
        try:
            self.subprotocol = self.response_headers["Sec-WebSocket-Protocol"]
        except KeyError:
            # If the server does not send the subprotocol header, using
            # the apollo subprotocol by default
            self.subprotocol = self.APOLLO_SUBPROTOCOL

        log.debug(f"backend subprotocol returned: {self.subprotocol!r}")

    async def _after_initialize(self):

        # If requested, create a task to send periodic pings to the backend
        if (
            self.subprotocol == self.GRAPHQLWS_SUBPROTOCOL
            and self.ping_interval is not None
        ):

            self.send_ping_task = asyncio.ensure_future(self._send_ping_coro())

    async def _close_hook(self):
        log.debug("_close_hook: start")

        # Properly shut down the send ping task if enabled
        if self.send_ping_task is not None:
            log.debug("_close_hook: cancelling send_ping_task")
            self.send_ping_task.cancel()
            with suppress(asyncio.CancelledError, TransportConnectionFailed):
                log.debug("_close_hook: awaiting send_ping_task")
                await self.send_ping_task
            self.send_ping_task = None

        log.debug("_close_hook: end")



import io
import json
import logging
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    List,
    NoReturn,
    Optional,
    Tuple,
    Type,
    Union,
)

import httpx
from graphql import ExecutionResult

from ..graphql_request import GraphQLRequest
from . import AsyncTransport, Transport
from .common.batch import get_batch_execution_result_list
from .exceptions import (
    TransportAlreadyConnected,
    TransportClosed,
    TransportConnectionFailed,
    TransportProtocolError,
    TransportServerError,
)
from .file_upload import close_files, extract_files, open_files

log = logging.getLogger(__name__)


class _HTTPXTransport:
    file_classes: Tuple[Type[Any], ...] = (io.IOBase,)

    response_headers: Optional[httpx.Headers] = None

    def __init__(
        self,
        url: Union[str, httpx.URL],
        json_serialize: Callable = json.dumps,
        json_deserialize: Callable = json.loads,
        **kwargs: Any,
    ):
        """Initialize the transport with the given httpx parameters.

        :param url: The GraphQL server URL. Example: 'https://server.com:PORT/path'.
        :param json_serialize: Json serializer callable.
                By default json.dumps() function.
        :param json_deserialize: Json deserializer callable.
                By default json.loads() function.
        :param kwargs: Extra args passed to the `httpx` client.
        """
        self.url = url
        self.json_serialize = json_serialize
        self.json_deserialize = json_deserialize
        self.kwargs = kwargs

    def _prepare_request(
        self,
        request: Union[GraphQLRequest, List[GraphQLRequest]],
        *,
        extra_args: Optional[Dict[str, Any]] = None,
        upload_files: bool = False,
    ) -> Dict[str, Any]:

        payload: Dict | List
        if isinstance(request, GraphQLRequest):
            payload = request.payload
        else:
            payload = [req.payload for req in request]

        if upload_files:
            assert isinstance(payload, Dict)
            assert isinstance(request, GraphQLRequest)
            post_args = self._prepare_file_uploads(request, payload)
        else:
            post_args = {"json": payload}

        # Log the payload
        if log.isEnabledFor(logging.DEBUG):
            log.debug(">>> %s", self.json_serialize(payload))

        # Pass post_args to httpx post method
        if extra_args:
            post_args.update(extra_args)

        return post_args

    def _prepare_file_uploads(
        self,
        request: GraphQLRequest,
        payload: Dict[str, Any],
    ) -> Dict[str, Any]:

        variable_values = request.variable_values

        # If the upload_files flag is set, then we need variable_values
        assert variable_values is not None

        # If we upload files, we will extract the files present in the
        # variable_values dict and replace them by null values
        nulled_variable_values, files = extract_files(
            variables=variable_values,
            file_classes=self.file_classes,
        )

        # Opening the files using the FileVar parameters
        open_files(list(files.values()))
        self.files = files

        # Save the nulled variable values in the payload
        payload["variables"] = nulled_variable_values

        # Prepare to send multipart-encoded data
        data: Dict[str, Any] = {}
        file_map: Dict[str, List[str]] = {}
        file_streams: Dict[str, Tuple[str, ...]] = {}

        for i, (path, file_var) in enumerate(files.items()):
            key = str(i)

            # Generate the file map
            # path is nested in a list because the spec allows multiple pointers
            # to the same file. But we don't support that.
            # Will generate something like {"0": ["variables.file"]}
            file_map[key] = [path]

            name = key if file_var.filename is None else file_var.filename

            if file_var.content_type is None:
                file_streams[key] = (name, file_var.f)
            else:
                file_streams[key] = (name, file_var.f, file_var.content_type)

        # Add the payload to the operations field
        operations_str = self.json_serialize(payload)
        log.debug("operations %s", operations_str)
        data["operations"] = operations_str

        # Add the file map field
        file_map_str = self.json_serialize(file_map)
        log.debug("file_map %s", file_map_str)
        data["map"] = file_map_str

        return {"data": data, "files": file_streams}

    def _get_json_result(self, response: httpx.Response) -> Any:

        # Saving latest response headers in the transport
        self.response_headers = response.headers

        if log.isEnabledFor(logging.DEBUG):
            log.debug("<<< %s", response.text)

        try:
            result: Dict[str, Any] = self.json_deserialize(response.content)
        except Exception:
            self._raise_response_error(response, "Not a JSON answer")

        return result

    def _prepare_result(self, response: httpx.Response) -> ExecutionResult:

        result = self._get_json_result(response)

        if "errors" not in result and "data" not in result:
            self._raise_response_error(response, 'No "data" or "errors" keys in answer')

        return ExecutionResult(
            errors=result.get("errors"),
            data=result.get("data"),
            extensions=result.get("extensions"),
        )

    def _prepare_batch_result(
        self,
        reqs: List[GraphQLRequest],
        response: httpx.Response,
    ) -> List[ExecutionResult]:

        answers = self._get_json_result(response)

        try:
            return get_batch_execution_result_list(reqs, answers)
        except TransportProtocolError:
            # Raise a TransportServerError if status > 400
            self._raise_transport_server_error_if_status_more_than_400(response)
            # In other cases, raise a TransportProtocolError
            raise

    @staticmethod
    def _raise_transport_server_error_if_status_more_than_400(
        response: httpx.Response,
    ) -> None:
        # If the status is >400,
        # then we need to raise a TransportServerError
        try:
            # Raise a HTTPStatusError if response status is 400 or higher
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            raise TransportServerError(str(e), e.response.status_code) from e

    @classmethod
    def _raise_response_error(cls, response: httpx.Response, reason: str) -> NoReturn:
        # We raise a TransportServerError if the status code is 400 or higher
        # We raise a TransportProtocolError in the other cases

        cls._raise_transport_server_error_if_status_more_than_400(response)

        raise TransportProtocolError(
            f"Server did not return a GraphQL result: " f"{reason}: " f"{response.text}"
        )


class HTTPXTransport(Transport, _HTTPXTransport):
    """:ref:`Sync Transport <sync_transports>` used to execute GraphQL queries
    on remote servers.

    The transport uses the httpx library to send HTTP POST requests.
    """

    client: Optional[httpx.Client] = None

    def connect(self):
        if self.client:
            raise TransportAlreadyConnected("Transport is already connected")

        log.debug("Connecting transport")

        self.client = httpx.Client(**self.kwargs)

    def execute(
        self,
        request: GraphQLRequest,
        *,
        extra_args: Optional[Dict[str, Any]] = None,
        upload_files: bool = False,
    ) -> ExecutionResult:
        """Execute GraphQL query.

        Execute the provided request against the configured remote server. This
        uses the httpx library to perform a HTTP POST request to the remote server.

        :param request: GraphQL request as a
                        :class:`GraphQLRequest <gql.GraphQLRequest>` object.
        :param extra_args: additional arguments to send to the httpx post method
        :param upload_files: Set to True if you want to put files in the variable values
        :return: The result of execution.
            `data` is the result of executing the query, `errors` is null
            if no errors occurred, and is a non-empty array if an error occurred.
        """
        if not self.client:
            raise TransportClosed("Transport is not connected")

        post_args = self._prepare_request(
            request,
            extra_args=extra_args,
            upload_files=upload_files,
        )

        try:
            response = self.client.post(self.url, **post_args)
        except Exception as e:
            raise TransportConnectionFailed(str(e)) from e
        finally:
            if upload_files:
                close_files(list(self.files.values()))

        return self._prepare_result(response)

    def execute_batch(
        self,
        reqs: List[GraphQLRequest],
        extra_args: Optional[Dict[str, Any]] = None,
    ) -> List[ExecutionResult]:
        """Execute multiple GraphQL requests in a batch.

        Don't call this coroutine directly on the transport, instead use
        :code:`execute_batch` on a client or a session.

        :param reqs: GraphQL requests as a list of GraphQLRequest objects.
        :param extra_args: additional arguments to send to the httpx post method
        :return: A list of results of execution.
            For every result `data` is the result of executing the query,
            `errors` is null if no errors occurred, and is a non-empty array
            if an error occurred.
        """

        if not self.client:
            raise TransportClosed("Transport is not connected")

        post_args = self._prepare_request(
            reqs,
            extra_args=extra_args,
        )

        try:
            response = self.client.post(self.url, **post_args)
        except Exception as e:
            raise TransportConnectionFailed(str(e)) from e

        return self._prepare_batch_result(reqs, response)

    def close(self):
        """Closing the transport by closing the inner session"""
        if self.client:
            self.client.close()
            self.client = None


class HTTPXAsyncTransport(AsyncTransport, _HTTPXTransport):
    """:ref:`Async Transport <async_transports>` used to execute GraphQL queries
    on remote servers.

    The transport uses the httpx library with anyio.
    """

    client: Optional[httpx.AsyncClient] = None

    async def connect(self):
        if self.client:
            raise TransportAlreadyConnected("Transport is already connected")

        log.debug("Connecting transport")

        self.client = httpx.AsyncClient(**self.kwargs)

    async def execute(
        self,
        request: GraphQLRequest,
        *,
        extra_args: Optional[Dict[str, Any]] = None,
        upload_files: bool = False,
    ) -> ExecutionResult:
        """Execute GraphQL query.

        Execute the provided request against the configured remote server. This
        uses the httpx library to perform a HTTP POST request asynchronously to the
        remote server.

        :param request: GraphQL request as a
                        :class:`GraphQLRequest <gql.GraphQLRequest>` object.
        :param extra_args: additional arguments to send to the httpx post method
        :param upload_files: Set to True if you want to put files in the variable values
        :return: The result of execution.
            `data` is the result of executing the query, `errors` is null
            if no errors occurred, and is a non-empty array if an error occurred.
        """
        if not self.client:
            raise TransportClosed("Transport is not connected")

        post_args = self._prepare_request(
            request,
            extra_args=extra_args,
            upload_files=upload_files,
        )

        try:
            response = await self.client.post(self.url, **post_args)
        except Exception as e:
            raise TransportConnectionFailed(str(e)) from e
        finally:
            if upload_files:
                close_files(list(self.files.values()))

        return self._prepare_result(response)

    async def execute_batch(
        self,
        reqs: List[GraphQLRequest],
        extra_args: Optional[Dict[str, Any]] = None,
    ) -> List[ExecutionResult]:
        """Execute multiple GraphQL requests in a batch.

        Don't call this coroutine directly on the transport, instead use
        :code:`execute_batch` on a client or a session.

        :param reqs: GraphQL requests as a list of GraphQLRequest objects.
        :param extra_args: additional arguments to send to the httpx post method
        :return: A list of results of execution.
            For every result `data` is the result of executing the query,
            `errors` is null if no errors occurred, and is a non-empty array
            if an error occurred.
        """

        if not self.client:
            raise TransportClosed("Transport is not connected")

        post_args = self._prepare_request(
            reqs,
            extra_args=extra_args,
        )

        try:
            response = await self.client.post(self.url, **post_args)
        except Exception as e:
            raise TransportConnectionFailed(str(e)) from e

        return self._prepare_batch_result(reqs, response)

    def subscribe(
        self,
        request: GraphQLRequest,
    ) -> AsyncGenerator[ExecutionResult, None]:
        """Subscribe is not supported on HTTP.

        :meta private:
        """
        raise NotImplementedError("The HTTP transport does not support subscriptions")

    async def close(self):
        """Closing the transport by closing the inner session"""
        if self.client:
            await self.client.aclose()
            self.client = None



import asyncio
from math import isfinite
from typing import Any, Dict, NamedTuple, Optional

import pytest
from graphql import ExecutionResult, graphql_sync
from graphql.error import GraphQLError
from graphql.language import ValueNode
from graphql.pyutils import inspect
from graphql.type import (
    GraphQLArgument,
    GraphQLField,
    GraphQLFloat,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
)
from graphql.utilities import value_from_ast_untyped

from gql import Client, gql
from gql.transport.exceptions import TransportQueryError
from gql.utilities import serialize_value, update_schema_scalar, update_schema_scalars

from ..conftest import MS

# Marking all tests in this file with the aiohttp marker
pytestmark = pytest.mark.aiohttp


class Money(NamedTuple):
    amount: float
    currency: str


def is_finite(value: Any) -> bool:
    """Return true if a value is a finite number."""
    return (isinstance(value, int) and not isinstance(value, bool)) or (
        isinstance(value, float) and isfinite(value)
    )


def serialize_money(output_value: Any) -> Dict[str, Any]:
    if not isinstance(output_value, Money):
        raise GraphQLError("Cannot serialize money value: " + inspect(output_value))
    return output_value._asdict()


def parse_money_value(input_value: Any) -> Money:
    """Using Money custom scalar from graphql-core tests except here the
    input value is supposed to be a dict instead of a Money object."""

    """
    if isinstance(input_value, Money):
        return input_value
    """

    if isinstance(input_value, dict):
        amount = input_value.get("amount", None)
        currency = input_value.get("currency", None)

        if not is_finite(amount) or not isinstance(currency, str):
            raise GraphQLError("Cannot parse money value dict: " + inspect(input_value))

        return Money(float(amount), currency)
    else:
        raise GraphQLError("Cannot parse money value: " + inspect(input_value))


def parse_money_literal(
    value_node: ValueNode, variables: Optional[Dict[str, Any]] = None
) -> Money:
    money = value_from_ast_untyped(value_node, variables)
    if variables is not None and (
        # variables are not set when checked with ValuesIOfCorrectTypeRule
        not money
        or not is_finite(money.get("amount"))
        or not isinstance(money.get("currency"), str)
    ):
        raise GraphQLError("Cannot parse literal money value: " + inspect(money))
    return Money(**money)


MoneyScalar = GraphQLScalarType(
    name="Money",
    serialize=serialize_money,
    parse_value=parse_money_value,
    parse_literal=parse_money_literal,
)

root_value = {
    "balance": Money(42, "DM"),
    "friends_balance": [Money(12, "EUR"), Money(24, "EUR"), Money(150, "DM")],
    "countries_balance": {
        "Belgium": Money(15000, "EUR"),
        "Luxembourg": Money(99999, "EUR"),
    },
}


def resolve_balance(root, _info):
    return root["balance"]


def resolve_friends_balance(root, _info):
    return root["friends_balance"]


def resolve_countries_balance(root, _info):
    return root["countries_balance"]


def resolve_belgium_balance(countries_balance, _info):
    return countries_balance["Belgium"]


def resolve_luxembourg_balance(countries_balance, _info):
    return countries_balance["Luxembourg"]


def resolve_to_euros(_root, _info, money):
    amount = money.amount
    currency = money.currency
    if not amount or currency == "EUR":
        return amount
    if currency == "DM":
        return amount * 0.5
    raise ValueError("Cannot convert to euros: " + inspect(money))


countriesBalance = GraphQLObjectType(
    name="CountriesBalance",
    fields={
        "Belgium": GraphQLField(
            GraphQLNonNull(MoneyScalar), resolve=resolve_belgium_balance
        ),
        "Luxembourg": GraphQLField(
            GraphQLNonNull(MoneyScalar), resolve=resolve_luxembourg_balance
        ),
    },
)

queryType = GraphQLObjectType(
    name="RootQueryType",
    fields={
        "balance": GraphQLField(MoneyScalar, resolve=resolve_balance),
        "toEuros": GraphQLField(
            GraphQLFloat,
            args={"money": GraphQLArgument(MoneyScalar)},
            resolve=resolve_to_euros,
        ),
        "friends_balance": GraphQLField(
            GraphQLList(MoneyScalar), resolve=resolve_friends_balance
        ),
        "countries_balance": GraphQLField(
            GraphQLNonNull(countriesBalance),
            resolve=resolve_countries_balance,
        ),
    },
)


def resolve_spent_money(spent_money, _info, **kwargs):
    return spent_money


async def subscribe_spend_all(_root, _info, money):
    while money.amount > 0:
        money = Money(money.amount - 1, money.currency)
        yield money
        await asyncio.sleep(1 * MS)


subscriptionType = GraphQLObjectType(
    "Subscription",
    fields=lambda: {
        "spend": GraphQLField(
            MoneyScalar,
            args={"money": GraphQLArgument(MoneyScalar)},
            subscribe=subscribe_spend_all,
            resolve=resolve_spent_money,
        )
    },
)

schema = GraphQLSchema(
    query=queryType,
    subscription=subscriptionType,
)


def test_custom_scalar_in_output():

    client = Client(schema=schema, parse_results=True)

    query = gql("{balance}")

    result = client.execute(query, root_value=root_value)

    print(result)

    assert result["balance"] == root_value["balance"]


def test_custom_scalar_in_output_embedded_fragments():

    client = Client(schema=schema, parse_results=True)

    query = gql(
        """
        fragment LuxMoneyInternal on CountriesBalance {
            ... on CountriesBalance {
                Luxembourg
            }
        }
        query {
            countries_balance {
                Belgium
                ...LuxMoney
            }
        }
        fragment LuxMoney on CountriesBalance {
            ...LuxMoneyInternal
        }
        """
    )

    result = client.execute(query, root_value=root_value)

    print(result)

    belgium_money = result["countries_balance"]["Belgium"]
    assert belgium_money == Money(15000, "EUR")
    luxembourg_money = result["countries_balance"]["Luxembourg"]
    assert luxembourg_money == Money(99999, "EUR")


def test_custom_scalar_list_in_output():

    client = Client(schema=schema, parse_results=True)

    query = gql("{friends_balance}")

    result = client.execute(query, root_value=root_value)

    print(result)

    assert result["friends_balance"] == root_value["friends_balance"]


def test_custom_scalar_in_input_query():

    client = Client(schema=schema)

    query = gql('{toEuros(money: {amount: 10, currency: "DM"})}')

    result = client.execute(query, root_value=root_value)

    assert result["toEuros"] == 5

    query = gql('{toEuros(money: {amount: 10, currency: "EUR"})}')

    result = client.execute(query, root_value=root_value)

    assert result["toEuros"] == 10


def test_custom_scalar_in_input_variable_values():

    client = Client(schema=schema)

    query = gql("query myquery($money: Money) {toEuros(money: $money)}")

    money_value = {"amount": 10, "currency": "DM"}

    query.variable_values = {"money": money_value}

    result = client.execute(query, root_value=root_value)

    assert result["toEuros"] == 5


def test_custom_scalar_in_input_variable_values_serialized():

    client = Client(schema=schema)

    query = gql("query myquery($money: Money) {toEuros(money: $money)}")

    money_value = Money(10, "DM")

    query.variable_values = {"money": money_value}

    result = client.execute(
        query,
        root_value=root_value,
        serialize_variables=True,
    )

    assert result["toEuros"] == 5


def test_custom_scalar_in_input_variable_values_serialized_with_operation_name():

    client = Client(schema=schema)

    query = gql("query myquery($money: Money) {toEuros(money: $money)}")

    money_value = Money(10, "DM")

    query.variable_values = {"money": money_value}
    query.operation_name = "myquery"

    result = client.execute(
        query,
        root_value=root_value,
        serialize_variables=True,
    )

    assert result["toEuros"] == 5


def test_serialize_variable_values_exception_multiple_ops_without_operation_name():

    client = Client(schema=schema)

    query = gql(
        """
    query myconversion($money: Money) {
        toEuros(money: $money)
    }

    query mybalance {
        balance
    }"""
    )

    money_value = Money(10, "DM")

    query.variable_values = {"money": money_value}

    with pytest.raises(GraphQLError) as exc_info:
        client.execute(
            query,
            root_value=root_value,
            serialize_variables=True,
        )

    exception = exc_info.value

    assert (
        str(exception)
        == "Must provide operation name if query contains multiple operations."
    )


def test_serialize_variable_values_exception_operation_name_not_found():

    client = Client(schema=schema)

    query = gql(
        """
    query myconversion($money: Money) {
        toEuros(money: $money)
    }
"""
    )

    money_value = Money(10, "DM")

    query.variable_values = {"money": money_value}
    query.operation_name = "invalid_operation_name"

    with pytest.raises(GraphQLError) as exc_info:
        client.execute(
            query,
            root_value=root_value,
            serialize_variables=True,
        )

    exception = exc_info.value

    assert str(exception) == "Unknown operation named 'invalid_operation_name'."


def test_custom_scalar_subscribe_in_input_variable_values_serialized():

    client = Client(schema=schema)

    query = gql("subscription spendAll($money: Money) {spend(money: $money)}")

    money_value = Money(10, "DM")

    query.variable_values = {"money": money_value}

    expected_result = {"spend": Money(10, "DM")}

    for result in client.subscribe(
        query,
        root_value=root_value,
        serialize_variables=True,
        parse_result=True,
    ):
        print(f"result = {result!r}")
        assert isinstance(result["spend"], Money)
        expected_result["spend"] = Money(expected_result["spend"].amount - 1, "DM")
        assert expected_result == result


async def make_money_backend(aiohttp_server):
    from aiohttp import web

    async def handler(request):
        req_data = await request.json()

        def handle_single(data: Dict[str, Any]) -> ExecutionResult:
            source = data["query"]
            try:
                variables = data["variables"]
            except KeyError:
                variables = None

            result = graphql_sync(
                schema, source, variable_values=variables, root_value=root_value
            )

            return result

        if isinstance(req_data, list):
            results = [handle_single(d) for d in req_data]

            return web.json_response(
                [
                    {
                        "data": result.data,
                        "errors": (
                            [str(e) for e in result.errors] if result.errors else None
                        ),
                    }
                    for result in results
                ]
            )
        else:
            result = handle_single(req_data)
            return web.json_response(
                {
                    "data": result.data,
                    "errors": (
                        [str(e) for e in result.errors] if result.errors else None
                    ),
                }
            )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    return server


async def make_money_transport(aiohttp_server):
    from gql.transport.aiohttp import AIOHTTPTransport

    server = await make_money_backend(aiohttp_server)

    url = server.make_url("/")

    transport = AIOHTTPTransport(url=url, timeout=10)

    return transport


async def make_sync_money_transport(aiohttp_server):
    from gql.transport.requests import RequestsHTTPTransport

    server = await make_money_backend(aiohttp_server)

    url = server.make_url("/")

    transport = RequestsHTTPTransport(url=url, timeout=10)

    return (server, transport)


@pytest.mark.asyncio
async def test_custom_scalar_in_output_with_transport(aiohttp_server):

    transport = await make_money_transport(aiohttp_server)

    async with Client(
        transport=transport,
    ) as session:

        query = gql("{balance}")

        result = await session.execute(query)

        print(result)

        assert result["balance"] == serialize_money(root_value["balance"])


@pytest.mark.asyncio
async def test_custom_scalar_in_input_query_with_transport(aiohttp_server):

    transport = await make_money_transport(aiohttp_server)

    async with Client(
        transport=transport,
    ) as session:

        query = gql('{toEuros(money: {amount: 10, currency: "DM"})}')

        result = await session.execute(query)

        assert result["toEuros"] == 5

        query = gql('{toEuros(money: {amount: 10, currency: "EUR"})}')

        result = await session.execute(query)

        assert result["toEuros"] == 10


@pytest.mark.asyncio
async def test_custom_scalar_in_input_variable_values_with_transport(aiohttp_server):

    transport = await make_money_transport(aiohttp_server)

    async with Client(
        transport=transport,
    ) as session:

        query = gql("query myquery($money: Money) {toEuros(money: $money)}")

        money_value = {"amount": 10, "currency": "DM"}
        # money_value = Money(10, "DM")

        query.variable_values = {"money": money_value}

        result = await session.execute(query)

        print(f"result = {result!r}")
        assert result["toEuros"] == 5


@pytest.mark.asyncio
async def test_custom_scalar_in_input_variable_values_split_with_transport(
    aiohttp_server,
):

    transport = await make_money_transport(aiohttp_server)

    async with Client(
        transport=transport,
    ) as session:

        query = gql(
            """
query myquery($amount: Float, $currency: String) {
    toEuros(money: {amount: $amount, currency: $currency})
}"""
        )

        query.variable_values = {"amount": 10, "currency": "DM"}

        result = await session.execute(query)

        print(f"result = {result!r}")
        assert result["toEuros"] == 5


@pytest.mark.asyncio
async def test_custom_scalar_serialize_variables(aiohttp_server):

    transport = await make_money_transport(aiohttp_server)

    async with Client(
        schema=schema,
        transport=transport,
    ) as session:

        query = gql("query myquery($money: Money) {toEuros(money: $money)}")

        query.variable_values = {"money": Money(10, "DM")}

        result = await session.execute(query, serialize_variables=True)

        print(f"result = {result!r}")
        assert result["toEuros"] == 5


@pytest.mark.asyncio
async def test_custom_scalar_serialize_variables_no_schema(aiohttp_server):

    transport = await make_money_transport(aiohttp_server)

    async with Client(
        transport=transport,
    ) as session:

        query = gql("query myquery($money: Money) {toEuros(money: $money)}")

        query.variable_values = {"money": Money(10, "DM")}

        with pytest.raises(TransportQueryError):
            await session.execute(query, serialize_variables=True)


@pytest.mark.asyncio
async def test_custom_scalar_serialize_variables_schema_from_introspection(
    aiohttp_server,
):

    transport = await make_money_transport(aiohttp_server)

    async with Client(transport=transport, fetch_schema_from_transport=True) as session:

        schema = session.client.schema

        # Updating the Money Scalar in the schema
        # We cannot replace it because some other objects keep a reference
        # to the existing Scalar
        # cannot do: schema.type_map["Money"] = MoneyScalar

        money_scalar = schema.type_map["Money"]

        money_scalar.serialize = MoneyScalar.serialize
        money_scalar.parse_value = MoneyScalar.parse_value
        money_scalar.parse_literal = MoneyScalar.parse_literal

        query = gql("query myquery($money: Money) {toEuros(money: $money)}")

        query.variable_values = {"money": Money(10, "DM")}

        result = await session.execute(query, serialize_variables=True)

        print(f"result = {result!r}")
        assert result["toEuros"] == 5


@pytest.mark.asyncio
async def test_update_schema_scalars(aiohttp_server):

    transport = await make_money_transport(aiohttp_server)

    async with Client(transport=transport, fetch_schema_from_transport=True) as session:

        # Update the schema MoneyScalar default implementation from
        # introspection with our provided conversion methods
        # update_schema_scalars(session.client.schema, [MoneyScalar])
        update_schema_scalar(session.client.schema, "Money", MoneyScalar)

        query = gql("query myquery($money: Money) {toEuros(money: $money)}")

        query.variable_values = {"money": Money(10, "DM")}

        result = await session.execute(query, serialize_variables=True)

        print(f"result = {result!r}")
        assert result["toEuros"] == 5


def test_update_schema_scalars_invalid_scalar():

    with pytest.raises(TypeError) as exc_info:
        update_schema_scalars(schema, [int])  # type: ignore

    exception = exc_info.value

    assert str(exception) == "Scalars should be instances of GraphQLScalarType."

    with pytest.raises(TypeError) as exc_info:
        update_schema_scalar(schema, "test", int)  # type: ignore

    exception = exc_info.value

    assert str(exception) == "Scalars should be instances of GraphQLScalarType."


def test_update_schema_scalars_invalid_scalar_argument():

    with pytest.raises(TypeError) as exc_info:
        update_schema_scalars(schema, MoneyScalar)  # type: ignore

    exception = exc_info.value

    assert str(exception) == "Scalars argument should be a list of scalars."


def test_update_schema_scalars_scalar_not_found_in_schema():

    NotFoundScalar = GraphQLScalarType(
        name="abcd",
    )

    with pytest.raises(KeyError) as exc_info:
        update_schema_scalars(schema, [MoneyScalar, NotFoundScalar])

    exception = exc_info.value

    assert "Scalar 'abcd' not found in schema." in str(exception)


def test_update_schema_scalars_scalar_type_is_not_a_scalar_in_schema():

    with pytest.raises(TypeError) as exc_info:
        update_schema_scalar(schema, "CountriesBalance", MoneyScalar)

    exception = exc_info.value

    assert 'The type "CountriesBalance" is not a GraphQLScalarType, it is a' in str(
        exception
    )


@pytest.mark.asyncio
@pytest.mark.requests
async def test_custom_scalar_serialize_variables_sync_transport(
    aiohttp_server, run_sync_test
):

    server, transport = await make_sync_money_transport(aiohttp_server)

    def test_code():
        with Client(schema=schema, transport=transport, parse_results=True) as session:

            query = gql("query myquery($money: Money) {toEuros(money: $money)}")

            query.variable_values = {"money": Money(10, "DM")}

            result = session.execute(query, serialize_variables=True)

            print(f"result = {result!r}")
            assert result["toEuros"] == 5

    await run_sync_test(server, test_code)


@pytest.mark.asyncio
@pytest.mark.requests
async def test_custom_scalar_serialize_variables_sync_transport_2(
    aiohttp_server, run_sync_test
):
    server, transport = await make_sync_money_transport(aiohttp_server)

    def test_code():
        with Client(schema=schema, transport=transport, parse_results=True) as session:

            query = gql("query myquery($money: Money) {toEuros(money: $money)}")

            query.variable_values = {"money": Money(10, "DM")}

            results = session.execute_batch(
                [
                    query,
                    query,
                ],
                serialize_variables=True,
            )

            print(f"result = {results!r}")
            assert results[0]["toEuros"] == 5
            assert results[1]["toEuros"] == 5

    await run_sync_test(server, test_code)


@pytest.mark.asyncio
@pytest.mark.aiohttp
async def test_custom_scalar_serialize_variables_async_transport(aiohttp_server):
    transport = await make_money_transport(aiohttp_server)

    async with Client(
        schema=schema, transport=transport, parse_results=True
    ) as session:

        query = gql("query myquery($money: Money) {toEuros(money: $money)}")

        query.variable_values = {"money": Money(10, "DM")}

        results = await session.execute_batch(
            [
                query,
                query,
            ],
            serialize_variables=True,
        )

        print(f"result = {results!r}")
        assert results[0]["toEuros"] == 5
        assert results[1]["toEuros"] == 5


def test_serialize_value_with_invalid_type():

    with pytest.raises(GraphQLError) as exc_info:
        serialize_value("Not a valid type", 50)  # type: ignore

    exception = exc_info.value

    assert (
        str(exception) == "Impossible to serialize value with type: 'Not a valid type'."
    )


def test_serialize_value_with_non_null_type_null():

    non_null_int = GraphQLNonNull(GraphQLInt)

    with pytest.raises(GraphQLError) as exc_info:
        serialize_value(non_null_int, None)

    exception = exc_info.value

    assert str(exception) == "Type Int! Cannot be None."


def test_serialize_value_with_nullable_type():

    nullable_int = GraphQLInt

    assert serialize_value(nullable_int, None) is None


@pytest.mark.asyncio
async def test_gql_cli_print_schema(aiohttp_server, capsys):

    from gql.cli import get_parser, main

    server = await make_money_backend(aiohttp_server)

    url = str(server.make_url("/"))

    parser = get_parser(with_examples=True)
    args = parser.parse_args([url, "--print-schema"])

    exit_code = await main(args)

    assert exit_code == 0

    # Check that the result has been printed on stdout
    captured = capsys.readouterr()
    captured_out = str(captured.out).strip()

    print(captured_out)
    assert (
        """
type Subscription {
  spend(money: Money): Money
}
""".strip()
        in captured_out
    )



import asyncio
from typing import Optional, Tuple

from graphql import ExecutionResult

ParsedAnswer = Tuple[str, Optional[ExecutionResult]]


class ListenerQueue:
    """Special queue used for each query waiting for server answers

    If the server is stopped while the listener is still waiting,
    Then we send an exception to the queue and this exception will be raised
    to the consumer once all the previous messages have been consumed from the queue
    """

    def __init__(self, query_id: int, send_stop: bool) -> None:
        self.query_id: int = query_id
        self.send_stop: bool = send_stop
        self._queue: asyncio.Queue = asyncio.Queue()
        self._closed: bool = False

    async def get(self) -> ParsedAnswer:

        try:
            item = self._queue.get_nowait()
        except asyncio.QueueEmpty:
            item = await self._queue.get()

        self._queue.task_done()

        # If we receive an exception when reading the queue, we raise it
        if isinstance(item, Exception):
            self._closed = True
            raise item

        # Don't need to save new answers or
        # send the stop message if we already received the complete message
        answer_type, execution_result = item
        if answer_type == "complete":
            self.send_stop = False
            self._closed = True

        return item

    async def put(self, item: ParsedAnswer) -> None:

        if not self._closed:
            await self._queue.put(item)

    async def set_exception(self, exception: Exception) -> None:

        # Put the exception in the queue
        await self._queue.put(exception)

        # Don't need to send stop messages in case of error
        self.send_stop = False
        self._closed = True



import pytest
from graphql import GraphQLError

from gql import Client, gql
from tests.starwars.schema import StarWarsSchema


@pytest.fixture
def client():
    return Client(schema=StarWarsSchema)


def test_hero_name_query(client):
    query = gql(
        """
        query HeroNameQuery {
          hero {
            name
          }
        }
        """
    )
    expected = {"hero": {"name": "R2-D2"}}
    result = client.execute(query)
    assert result == expected


def test_hero_name_and_friends_query(client):
    query = gql(
        """
        query HeroNameAndFriendsQuery {
          hero {
            id
            name
            friends {
              name
            }
          }
        }
        """
    )
    expected = {
        "hero": {
            "id": "2001",
            "name": "R2-D2",
            "friends": [
                {"name": "Luke Skywalker"},
                {"name": "Han Solo"},
                {"name": "Leia Organa"},
            ],
        }
    }
    result = client.execute(query)
    assert result == expected


def test_nested_query(client):
    query = gql(
        """
        query NestedQuery {
          hero {
            name
            friends {
              name
              appearsIn
              friends {
                name
              }
            }
          }
        }
        """
    )
    expected = {
        "hero": {
            "name": "R2-D2",
            "friends": [
                {
                    "name": "Luke Skywalker",
                    "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                    "friends": [
                        {"name": "Han Solo"},
                        {"name": "Leia Organa"},
                        {"name": "C-3PO"},
                        {"name": "R2-D2"},
                    ],
                },
                {
                    "name": "Han Solo",
                    "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                    "friends": [
                        {"name": "Luke Skywalker"},
                        {"name": "Leia Organa"},
                        {"name": "R2-D2"},
                    ],
                },
                {
                    "name": "Leia Organa",
                    "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                    "friends": [
                        {"name": "Luke Skywalker"},
                        {"name": "Han Solo"},
                        {"name": "C-3PO"},
                        {"name": "R2-D2"},
                    ],
                },
            ],
        }
    }
    result = client.execute(query, parse_result=False)
    assert result == expected


def test_fetch_luke_query(client):
    query = gql(
        """
        query FetchLukeQuery {
          human(id: "1000") {
            name
          }
        }
        """
    )
    expected = {"human": {"name": "Luke Skywalker"}}
    result = client.execute(query)
    assert result == expected


def test_fetch_some_id_query(client):
    query = gql(
        """
        query FetchSomeIDQuery($someId: String!) {
          human(id: $someId) {
            name
          }
        }
        """
    )
    query.variable_values = {
        "someId": "1000",
    }
    expected = {"human": {"name": "Luke Skywalker"}}
    result = client.execute(query)
    assert result == expected


def test_fetch_some_id_query2(client):
    query = gql(
        """
        query FetchSomeIDQuery($someId: String!) {
          human(id: $someId) {
            name
          }
        }
        """
    )
    query.variable_values = {
        "someId": "1002",
    }
    expected = {"human": {"name": "Han Solo"}}
    result = client.execute(query)
    assert result == expected


def test_invalid_id_query(client):
    query = gql(
        """
        query humanQuery($id: String!) {
          human(id: $id) {
            name
          }
        }
        """
    )
    query.variable_values = {
        "id": "not a valid id",
    }
    expected = {"human": None}
    result = client.execute(query)
    assert result == expected


def test_fetch_luke_aliased(client):
    query = gql(
        """
        query FetchLukeAliased {
          luke: human(id: "1000") {
            name
          }
        }
        """
    )
    expected = {"luke": {"name": "Luke Skywalker"}}
    result = client.execute(query)
    assert result == expected


def test_fetch_luke_and_leia_aliased(client):
    query = gql(
        """
        query FetchLukeAndLeiaAliased {
          luke: human(id: "1000") {
            name
          }
          leia: human(id: "1003") {
            name
          }
        }
        """
    )
    expected = {"luke": {"name": "Luke Skywalker"}, "leia": {"name": "Leia Organa"}}
    result = client.execute(query)
    assert result == expected


def test_duplicate_fields(client):
    query = gql(
        """
        query DuplicateFields {
          luke: human(id: "1000") {
            name
            homePlanet
          }
          leia: human(id: "1003") {
            name
            homePlanet
          }
        }
        """
    )
    expected = {
        "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
        "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
    }
    result = client.execute(query)
    assert result == expected


def test_use_fragment(client):
    query = gql(
        """
        query UseFragment {
          luke: human(id: "1000") {
            ...HumanFragment
          }
          leia: human(id: "1003") {
            ...HumanFragment
          }
        }
        fragment HumanFragment on Human {
          name
          homePlanet
        }
        """
    )
    expected = {
        "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
        "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
    }
    result = client.execute(query)
    assert result == expected


def test_check_type_of_r2(client):
    query = gql(
        """
        query CheckTypeOfR2 {
          hero {
            __typename
            name
          }
        }
        """
    )
    expected = {"hero": {"__typename": "Droid", "name": "R2-D2"}}
    result = client.execute(query)
    assert result == expected


def test_check_type_of_luke(client):
    query = gql(
        """
        query CheckTypeOfLuke {
          hero(episode: EMPIRE) {
            __typename
            name
          }
        }
        """
    )
    expected = {"hero": {"__typename": "Human", "name": "Luke Skywalker"}}
    result = client.execute(query)
    assert result == expected


def test_parse_error(client):
    with pytest.raises(Exception) as exc_info:
        gql(
            """
            qeury
            """
        )
    error = exc_info.value
    assert isinstance(error, GraphQLError)
    assert "Syntax Error: Unexpected Name 'qeury'." in str(error)


def test_mutation_result(client):
    query = gql(
        """
        mutation CreateReviewForEpisode($ep: Episode!, $review: ReviewInput!) {
          createReview(episode: $ep, review: $review) {
            stars
            commentary
          }
        }
        """
    )
    query.variable_values = {
        "ep": "JEDI",
        "review": {"stars": 5, "commentary": "This is a great movie!"},
    }
    expected = {"createReview": {"stars": 5, "commentary": "This is a great movie!"}}
    result = client.execute(query)
    assert result == expected



from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport

# Select your transport with a defined url endpoint
transport = AIOHTTPTransport(url="https://countries.trevorblades.com/")

# Create a GraphQL client using the defined transport
client = Client(transport=transport, fetch_schema_from_transport=True)

# Provide a GraphQL query
query = gql(
    """
    query getContinents {
      continents {
        code
        name
      }
    }
"""
)

# Execute the query on the transport
result = client.execute(query)
print(result)



import asyncio

import pytest
from graphql import ExecutionResult, GraphQLError, subscribe

from gql import Client, gql

from .fixtures import reviews
from .schema import StarWarsSchema

subscription_str = """
    subscription ListenEpisodeReviews($ep: Episode!) {
      reviewAdded(episode: $ep) {
        stars,
        commentary,
        episode
      }
    }
"""


async def await_if_coroutine(obj):
    """Function to make tests work for graphql-core versions before and after 3.3.0a3"""
    if asyncio.iscoroutine(obj):
        return await obj

    return obj


@pytest.mark.asyncio
async def test_subscription_support():
    # reset review data for this test
    reviews[6] = [
        {"stars": 3, "commentary": "Was expecting more stuff", "episode": 6},
        {"stars": 5, "commentary": "This is a great movie!", "episode": 6},
    ]

    subs = gql(subscription_str)

    params = {"ep": "JEDI"}
    expected = [{**review, "episode": "JEDI"} for review in reviews[6]]

    ai = await await_if_coroutine(
        subscribe(StarWarsSchema, subs.document, variable_values=params)
    )

    result = [result.data["reviewAdded"] async for result in ai]

    assert result == expected


@pytest.mark.asyncio
async def test_subscription_support_using_client():
    # reset review data for this test
    reviews[6] = [
        {"stars": 3, "commentary": "Was expecting more stuff", "episode": 6},
        {"stars": 5, "commentary": "This is a great movie!", "episode": 6},
    ]

    subs = gql(subscription_str)

    subs.variable_values = {"ep": "JEDI"}
    expected = [{**review, "episode": "JEDI"} for review in reviews[6]]

    async with Client(schema=StarWarsSchema) as session:
        results = [
            result["reviewAdded"]
            async for result in await await_if_coroutine(
                session.subscribe(subs, parse_result=False)
            )
        ]

    assert results == expected


subscription_invalid_str = """
    subscription ListenEpisodeReviews($ep: Episode!) {
      qsdfqsdfqsdf
    }
"""


@pytest.mark.asyncio
async def test_subscription_support_using_client_invalid_field():

    subs = gql(subscription_invalid_str)

    subs.variable_values = {"ep": "JEDI"}

    async with Client(schema=StarWarsSchema) as session:

        # We subscribe directly from the transport to avoid local validation
        results = [
            result
            async for result in await await_if_coroutine(
                session.transport.subscribe(subs)
            )
        ]

    assert len(results) == 1
    result = results[0]
    assert isinstance(result, ExecutionResult)
    assert result.data is None
    assert isinstance(result.errors, list)
    assert len(result.errors) == 1
    error = result.errors[0]
    assert isinstance(error, GraphQLError)
    assert error.message == "The subscription field 'qsdfqsdfqsdf' is not defined."



import asyncio
import logging

import backoff

from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport

logging.basicConfig(level=logging.INFO)


async def main():

    # Note: this example used the test backend from
    # https://github.com/slothmanxyz/typegraphql-ws-apollo
    transport = AIOHTTPTransport(url="ws://localhost:5000/graphql")

    client = Client(transport=transport)

    retry_connect = backoff.on_exception(
        backoff.expo,
        Exception,
        max_value=10,
        jitter=None,
    )
    session = await client.connect_async(reconnecting=True, retry_connect=retry_connect)

    num = 0

    while True:
        num += 1

        # Execute single query
        query = gql("mutation ($message: String!) {sendMessage(message: $message)}")

        query.variable_values = {"message": f"test {num}"}

        try:
            result = await session.execute(query)
            print(result)
        except Exception as e:
            print(f"Received exception {e}")

        await asyncio.sleep(1)


asyncio.run(main())



import pytest


class FakeRequest:
    headers = None

    def __init__(self, request_props=None):
        if not isinstance(request_props, dict):
            return
        self.method = request_props.get("method")
        self.url = request_props.get("url")
        self.headers = request_props.get("headers")
        self.context = request_props.get("context")
        self.body = request_props.get("body")


@pytest.fixture
def fake_request_factory():
    def _fake_request_factory(request_props=None):
        return FakeRequest(request_props=request_props)

    yield _fake_request_factory



import asyncio
import json
import sys
import warnings
from typing import List

import pytest
from graphql import ExecutionResult
from parse import search

from gql import Client, gql
from gql.client import AsyncClientSession
from gql.transport.exceptions import TransportConnectionFailed, TransportServerError

from .conftest import MS, WebSocketServerHelper
from .starwars.schema import StarWarsIntrospection, StarWarsSchema, StarWarsTypeDef

# Marking all tests in this file with the aiohttp AND websockets marker
pytestmark = [pytest.mark.aiohttp, pytest.mark.websockets]

starwars_expected_one = {
    "stars": 3,
    "commentary": "Was expecting more stuff",
    "episode": "JEDI",
}

starwars_expected_two = {
    "stars": 5,
    "commentary": "This is a great movie!",
    "episode": "JEDI",
}


async def server_starwars(ws):
    import websockets

    await WebSocketServerHelper.send_connection_ack(ws)

    try:
        await ws.recv()

        reviews = [starwars_expected_one, starwars_expected_two]

        for review in reviews:

            data = (
                '{"type":"data","id":"1","payload":{"data":{"reviewAdded": '
                + json.dumps(review)
                + "}}}"
            )
            await ws.send(data)
            await asyncio.sleep(2 * MS)

        await WebSocketServerHelper.send_complete(ws, 1)
        await WebSocketServerHelper.wait_connection_terminate(ws)

    except websockets.exceptions.ConnectionClosedOK:
        pass

    print("Server is now closed")


starwars_subscription_str = """
    subscription ListenEpisodeReviews($ep: Episode!) {
      reviewAdded(episode: $ep) {
        stars,
        commentary,
        episode
      }
    }
"""

starwars_invalid_subscription_str = """
    subscription ListenEpisodeReviews($ep: Episode!) {
      reviewAdded(episode: $ep) {
        not_valid_field,
        stars,
        commentary,
        episode
      }
    }
"""

countdown_server_answer = (
    '{{"type":"data","id":"{query_id}","payload":{{"data":{{"number":{number}}}}}}}'
)

WITH_KEEPALIVE = False


# List which can used to store received messages by the server
logged_messages: List[str] = []


async def server_countdown(ws):
    import websockets

    logged_messages.clear()

    global WITH_KEEPALIVE
    try:
        await WebSocketServerHelper.send_connection_ack(ws)
        if WITH_KEEPALIVE:
            await WebSocketServerHelper.send_keepalive(ws)

        result = await ws.recv()
        logged_messages.append(result)

        json_result = json.loads(result)
        assert json_result["type"] == "start"
        payload = json_result["payload"]
        query = payload["query"]
        query_id = json_result["id"]

        count_found = search("count: {:d}", query)
        count = count_found[0]
        print(f"Countdown started from: {count}")

        async def counting_coro():
            for number in range(count, -1, -1):
                await ws.send(
                    countdown_server_answer.format(query_id=query_id, number=number)
                )
                await asyncio.sleep(2 * MS)

        counting_task = asyncio.ensure_future(counting_coro())

        async def stopping_coro():
            nonlocal counting_task
            while True:

                try:
                    result = await ws.recv()
                    logged_messages.append(result)
                except websockets.exceptions.ConnectionClosed:
                    break

                json_result = json.loads(result)

                if json_result["type"] == "stop" and json_result["id"] == str(query_id):
                    print("Cancelling counting task now")
                    counting_task.cancel()

        async def keepalive_coro():
            while True:
                await asyncio.sleep(5 * MS)
                try:
                    await WebSocketServerHelper.send_keepalive(ws)
                except websockets.exceptions.ConnectionClosed:
                    break

        stopping_task = asyncio.ensure_future(stopping_coro())
        if WITH_KEEPALIVE:
            keepalive_task = asyncio.ensure_future(keepalive_coro())

        try:
            await counting_task
        except asyncio.CancelledError:
            print("Now counting task is cancelled")
        except Exception as exc:
            print(f"Exception in counting task: {exc!s}")

        stopping_task.cancel()

        try:
            await stopping_task
        except asyncio.CancelledError:
            print("Now stopping task is cancelled")

        if WITH_KEEPALIVE:
            keepalive_task.cancel()
            try:
                await keepalive_task
            except asyncio.CancelledError:
                print("Now keepalive task is cancelled")

        await WebSocketServerHelper.send_complete(ws, query_id)
        await WebSocketServerHelper.wait_connection_terminate(ws)
    except websockets.exceptions.ConnectionClosedOK:
        pass
    finally:
        await ws.wait_closed()


countdown_subscription_str = """
    subscription {{
      countdown (count: {count}) {{
        number
      }}
    }}
"""


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription(
    aiohttp_client_and_server, subscription_str
):

    session, server = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async for result in session.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_get_execution_result(
    aiohttp_client_and_server, subscription_str
):

    session, server = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async for result in session.subscribe(subscription, get_execution_result=True):

        assert isinstance(result, ExecutionResult)
        assert result.data is not None

        number = result.data["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_break(
    aiohttp_client_and_server, subscription_str
):

    session, server = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    generator = session.subscribe(subscription)
    async for result in generator:

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count

        if count <= 5:
            break

        count -= 1

    assert count == 5

    # Using aclose here to make it stop cleanly on pypy
    await generator.aclose()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_task_cancel(
    aiohttp_client_and_server, subscription_str
):

    session, server = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    task_cancelled = False

    async def task_coro():
        nonlocal count
        nonlocal task_cancelled

        try:
            async for result in session.subscribe(subscription):

                number = result["number"]
                print(f"Number received: {number}")

                assert number == count

                count -= 1
        except asyncio.CancelledError:
            print("Inside task cancelled")
            task_cancelled = True

    task = asyncio.ensure_future(task_coro())

    async def cancel_task_coro():
        nonlocal task

        await asyncio.sleep(11 * MS)

        task.cancel()

    cancel_task = asyncio.ensure_future(cancel_task_coro())

    await asyncio.gather(task, cancel_task)

    assert count > 0
    assert task_cancelled is True


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_close_transport(
    aiohttp_client_and_server, subscription_str
):

    session, _ = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async def task_coro():
        nonlocal count
        async for result in session.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count

            count -= 1

    task = asyncio.ensure_future(task_coro())

    async def close_transport_task_coro():
        nonlocal task

        await asyncio.sleep(11 * MS)

        await session.transport.close()

    close_transport_task = asyncio.ensure_future(close_transport_task_coro())

    await asyncio.gather(task, close_transport_task)

    assert count > 0


async def server_countdown_close_connection_in_middle(ws):
    await WebSocketServerHelper.send_connection_ack(ws)

    result = await ws.recv()
    json_result = json.loads(result)
    assert json_result["type"] == "start"
    payload = json_result["payload"]
    query = payload["query"]
    query_id = json_result["id"]

    count_found = search("count: {:d}", query)
    count = count_found[0]
    stopping_before = count // 2
    print(f"Countdown started from: {count}, stopping server before {stopping_before}")
    for number in range(count, stopping_before, -1):
        await ws.send(countdown_server_answer.format(query_id=query_id, number=number))
        await asyncio.sleep(2 * MS)

    print("Closing server while subscription is still running now")
    await ws.close()
    await ws.wait_closed()
    print("Server is now closed")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server", [server_countdown_close_connection_in_middle], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_server_connection_closed(
    aiohttp_client_and_server, subscription_str
):

    session, server = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    with pytest.raises(TransportConnectionFailed):

        async for result in session.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count

            count -= 1


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_slow_consumer(
    aiohttp_client_and_server, subscription_str
):

    session, server = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async for result in session.subscribe(subscription):
        await asyncio.sleep(10 * MS)

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count

        count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_with_operation_name(
    aiohttp_client_and_server, subscription_str
):

    session, server = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))
    subscription.operation_name = "CountdownSubscription"

    async for result in session.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1

    # Check that the query contains the operationName
    assert '"operationName": "CountdownSubscription"' in logged_messages[0]


WITH_KEEPALIVE = True


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_with_keepalive(
    aiohttp_client_and_server, subscription_str
):

    session, server = aiohttp_client_and_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async for result in session.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_with_keepalive_with_timeout_ok(
    server, subscription_str
):

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = AIOHTTPWebsocketsTransport(url=url, keep_alive_timeout=(20 * MS))

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async with client as session:
        async for result in session.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count
            count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_with_keepalive_with_timeout_nok(
    server, subscription_str
):

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = AIOHTTPWebsocketsTransport(url=url, keep_alive_timeout=(1 * MS))

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async with client as session:
        with pytest.raises(TransportServerError) as exc_info:
            async for result in session.subscribe(subscription):

                number = result["number"]
                print(f"Number received: {number}")

                assert number == count
                count -= 1

        assert "No keep-alive message has been received" in str(exc_info.value)


@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
def test_aiohttp_websocket_subscription_sync(server, subscription_str):
    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(url=url)

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    for result in client.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1


@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
def test_aiohttp_websocket_subscription_sync_user_exception(server, subscription_str):
    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(url=url)

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    with pytest.raises(Exception) as exc_info:
        for result in client.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count
            count -= 1

            if count == 5:
                raise Exception("This is an user exception")

    assert count == 5
    assert "This is an user exception" in str(exc_info.value)


@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
def test_aiohttp_websocket_subscription_sync_break(server, subscription_str):
    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(url=url)

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    for result in client.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

        if count == 5:
            break

    assert count == 5


@pytest.mark.skipif(sys.platform.startswith("win"), reason="test failing on windows")
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
def test_aiohttp_websocket_subscription_sync_graceful_shutdown(
    server, subscription_str
):
    """Note: this test will simulate a control-C happening while a sync subscription
    is in progress. To do that we will throw a KeyboardInterrupt exception inside
    the subscription async generator.

    The code should then do a clean close:
      - send stop messages for each active query
      - send a connection_terminate message
    Then the KeyboardInterrupt will be reraise (to warn potential user code)

    This test does not work on Windows but the behaviour with Windows is correct.
    """
    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(url=url)

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    interrupt_task = None

    with pytest.raises(KeyboardInterrupt):
        for result in client.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count

            if count == 5:

                # Simulate a KeyboardInterrupt in the generator
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore", message="There is no current event loop"
                    )
                    assert isinstance(client.session, AsyncClientSession)
                    interrupt_task = asyncio.ensure_future(
                        client.session._generator.athrow(KeyboardInterrupt)
                    )

            count -= 1

    assert count == 4

    # Catch interrupt_task exception to remove warning
    assert interrupt_task is not None
    interrupt_task.exception()

    # Check that the server received a connection_terminate message last
    assert logged_messages.pop() == '{"type": "connection_terminate"}'


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_aiohttp_websocket_subscription_running_in_thread(
    server, subscription_str, run_sync_test
):
    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    def test_code():
        path = "/graphql"
        url = f"ws://{server.hostname}:{server.port}{path}"
        transport = AIOHTTPWebsocketsTransport(url=url)

        client = Client(transport=transport)

        count = 10
        subscription = gql(subscription_str.format(count=count))

        for result in client.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count
            count -= 1

        assert count == -1

    await run_sync_test(server, test_code)


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_starwars], indirect=True)
@pytest.mark.parametrize("subscription_str", [starwars_subscription_str])
@pytest.mark.parametrize(
    "client_params",
    [
        {"schema": StarWarsSchema},
        {"introspection": StarWarsIntrospection},
        {"schema": StarWarsTypeDef},
    ],
)
async def test_async_aiohttp_client_validation(server, subscription_str, client_params):

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"

    transport = AIOHTTPWebsocketsTransport(url=url)

    client = Client(transport=transport, **client_params)

    async with client as session:

        subscription = gql(subscription_str)

        subscription.variable_values = {"ep": "JEDI"}

        expected = []

        async for result in session.subscribe(subscription, parse_result=False):

            review = result["reviewAdded"]
            expected.append(review)

            assert "stars" in review
            assert "commentary" in review
            assert "episode" in review

        assert expected[0] == starwars_expected_one


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_subscribe_on_closing_transport(server, subscription_str):

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"

    transport = AIOHTTPWebsocketsTransport(url=url)

    client = Client(transport=transport)
    count = 1
    subscription = gql(subscription_str.format(count=count))

    async with client as session:
        session.transport.adapter.websocket._writer._closing = True

        with pytest.raises(TransportConnectionFailed):
            async for _ in session.subscribe(subscription):
                pass


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_subscribe_on_null_transport(server, subscription_str):

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"

    transport = AIOHTTPWebsocketsTransport(url=url)

    client = Client(transport=transport)
    count = 1
    subscription = gql(subscription_str.format(count=count))

    async with client as session:

        session.transport.adapter.websocket = None
        with pytest.raises(TransportConnectionFailed):
            async for _ in session.subscribe(subscription):
                pass



import asyncio
import logging
import time
import warnings
from concurrent.futures import Future
from queue import Queue
from threading import Event, Thread
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import backoff
from anyio import fail_after
from graphql import (
    ExecutionResult,
    GraphQLSchema,
    IntrospectionQuery,
    build_ast_schema,
    parse,
    validate,
)

from .graphql_request import GraphQLRequest, support_deprecated_request
from .transport.async_transport import AsyncTransport
from .transport.exceptions import TransportConnectionFailed, TransportQueryError
from .transport.local_schema import LocalSchemaTransport
from .transport.transport import Transport
from .utilities import build_client_schema, get_introspection_query_ast
from .utilities import parse_result as parse_result_fn
from .utils import str_first_element

log = logging.getLogger(__name__)


class Client:
    """The Client class is the main entrypoint to execute GraphQL requests
    on a GQL transport.

    It can take sync or async transports as argument and can either execute
    and subscribe to requests itself with the
    :func:`execute <gql.client.Client.execute>` and
    :func:`subscribe <gql.client.Client.subscribe>` methods
    OR can be used to get a sync or async session depending on the
    transport type.

    To connect to an :ref:`async transport <async_transports>` and get an
    :class:`async session <gql.client.AsyncClientSession>`,
    use :code:`async with client as session:`

    To connect to a :ref:`sync transport <sync_transports>` and get a
    :class:`sync session <gql.client.SyncClientSession>`,
    use :code:`with client as session:`
    """

    def __init__(
        self,
        *,
        schema: Optional[Union[str, GraphQLSchema]] = None,
        introspection: Optional[IntrospectionQuery] = None,
        transport: Optional[Union[Transport, AsyncTransport]] = None,
        fetch_schema_from_transport: bool = False,
        introspection_args: Optional[Dict] = None,
        execute_timeout: Optional[Union[int, float]] = 10,
        serialize_variables: bool = False,
        parse_results: bool = False,
        batch_interval: float = 0,
        batch_max: int = 10,
    ):
        """Initialize the client with the given parameters.

        :param schema: an optional GraphQL Schema for local validation
                See :ref:`schema_validation`
        :param transport: The provided :ref:`transport <Transports>`.
        :param fetch_schema_from_transport: Boolean to indicate that if we want to fetch
                the schema from the transport using an introspection query.
        :param introspection_args: arguments passed to the
                :meth:`gql.utilities.get_introspection_query_ast` method.
        :param execute_timeout: The maximum time in seconds for the execution of a
                request before a TimeoutError is raised. Only used for async transports.
                Passing None results in waiting forever for a response.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums. Default: False.
        :param parse_results: Whether gql will try to parse the serialized output
                sent by the backend. Can be used to deserialize custom scalars or enums.
        :param batch_interval: Time to wait in seconds for batching requests together.
                Batching is disabled (by default) if 0.
        :param batch_max: Maximum number of requests in a single batch.
        """

        if introspection:
            assert (
                not schema
            ), "Cannot provide introspection and schema at the same time."
            schema = build_client_schema(introspection)

        if isinstance(schema, str):
            type_def_ast = parse(schema)
            schema = build_ast_schema(type_def_ast)

        if transport and fetch_schema_from_transport:
            assert (
                not schema
            ), "Cannot fetch the schema from transport if is already provided."

            assert not type(transport).__name__ == "AppSyncWebsocketsTransport", (
                "fetch_schema_from_transport=True is not allowed "
                "for AppSyncWebsocketsTransport "
                "because only subscriptions are allowed on the realtime endpoint."
            )

        if schema and not transport:
            transport = LocalSchemaTransport(schema)

        # GraphQL schema
        self.schema: Optional[GraphQLSchema] = schema

        # Answer of the introspection query
        self.introspection: Optional[IntrospectionQuery] = introspection

        # GraphQL transport chosen
        assert (
            transport is not None
        ), "You need to provide either a transport or a schema to the Client."
        self.transport: Union[Transport, AsyncTransport] = transport

        # Flag to indicate that we need to fetch the schema from the transport
        # On async transports, we fetch the schema before executing the first query
        self.fetch_schema_from_transport: bool = fetch_schema_from_transport
        self.introspection_args = (
            {} if introspection_args is None else introspection_args
        )

        # Enforced timeout of the execute function (only for async transports)
        self.execute_timeout = execute_timeout

        self.serialize_variables = serialize_variables
        self.parse_results = parse_results
        self.batch_interval = batch_interval
        self.batch_max = batch_max

    @property
    def batching_enabled(self) -> bool:
        return self.batch_interval != 0

    def validate(self, request: GraphQLRequest) -> None:
        """:meta private:"""
        assert (
            self.schema
        ), "Cannot validate the document locally, you need to pass a schema."

        validation_errors = validate(self.schema, request.document)
        if validation_errors:
            raise validation_errors[0]

    def _build_schema_from_introspection(
        self, execution_result: ExecutionResult
    ) -> None:
        if execution_result.errors:
            raise TransportQueryError(
                (
                    "Error while fetching schema: "
                    f"{str_first_element(execution_result.errors)}\n"
                    "If you don't need the schema, you can try with: "
                    '"fetch_schema_from_transport=False"'
                ),
                errors=execution_result.errors,
                data=execution_result.data,
                extensions=execution_result.extensions,
            )

        self.introspection = cast(IntrospectionQuery, execution_result.data)
        self.schema = build_client_schema(self.introspection)

    @staticmethod
    def _get_event_loop() -> asyncio.AbstractEventLoop:
        """Get the current asyncio event loop.

        Or create a new event loop if there isn't one (in a new Thread).
        """
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", message="There is no current event loop"
                )
                loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        return loop

    @overload
    def execute_sync(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> Dict[str, Any]: ...  # pragma: no cover

    @overload
    def execute_sync(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> ExecutionResult: ...  # pragma: no cover

    @overload
    def execute_sync(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]: ...  # pragma: no cover

    def execute_sync(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]:
        """:meta private:"""
        with self as session:
            return session.execute(
                request,
                serialize_variables=serialize_variables,
                parse_result=parse_result,
                get_execution_result=get_execution_result,
                **kwargs,
            )

    @overload
    def execute_batch_sync(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> List[Dict[str, Any]]: ...  # pragma: no cover

    @overload
    def execute_batch_sync(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> List[ExecutionResult]: ...  # pragma: no cover

    @overload
    def execute_batch_sync(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]: ...  # pragma: no cover

    def execute_batch_sync(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]:
        """:meta private:"""
        with self as session:
            return session.execute_batch(
                requests,
                serialize_variables=serialize_variables,
                parse_result=parse_result,
                get_execution_result=get_execution_result,
                **kwargs,
            )

    @overload
    async def execute_async(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> Dict[str, Any]: ...  # pragma: no cover

    @overload
    async def execute_async(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> ExecutionResult: ...  # pragma: no cover

    @overload
    async def execute_async(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]: ...  # pragma: no cover

    async def execute_async(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]:
        """:meta private:"""
        async with self as session:
            return await session.execute(
                request,
                serialize_variables=serialize_variables,
                parse_result=parse_result,
                get_execution_result=get_execution_result,
                **kwargs,
            )

    @overload
    async def execute_batch_async(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> List[Dict[str, Any]]: ...  # pragma: no cover

    @overload
    async def execute_batch_async(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> List[ExecutionResult]: ...  # pragma: no cover

    @overload
    async def execute_batch_async(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]: ...  # pragma: no cover

    async def execute_batch_async(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]:
        """:meta private:"""
        async with self as session:
            return await session.execute_batch(
                requests,
                serialize_variables=serialize_variables,
                parse_result=parse_result,
                get_execution_result=get_execution_result,
                **kwargs,
            )

    @overload
    def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> Dict[str, Any]: ...  # pragma: no cover

    @overload
    def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> ExecutionResult: ...  # pragma: no cover

    @overload
    def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]: ...  # pragma: no cover

    def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]:
        """Execute the provided request against the remote server using
        the transport provided during init.

        This function **WILL BLOCK** until the result is received from the server.

        Either the transport is sync and we execute the query synchronously directly
        OR the transport is async and we execute the query in the asyncio loop
        (blocking here until answer).

        This method will:

         - connect using the transport to get a session
         - execute the GraphQL request on the transport session
         - close the session and close the connection to the server

         If you have multiple requests to send, it is better to get your own session
         and execute the requests in your session.

         The extra arguments passed in the method will be passed to the transport
         execute method.
        """

        if isinstance(self.transport, AsyncTransport):
            loop = self._get_event_loop()

            assert not loop.is_running(), (
                "Cannot run client.execute(query) if an asyncio loop is running."
                " Use 'await client.execute_async(query)' instead."
            )

            data = loop.run_until_complete(
                self.execute_async(
                    request,
                    serialize_variables=serialize_variables,
                    parse_result=parse_result,
                    get_execution_result=get_execution_result,
                    **kwargs,
                )
            )

            return data

        else:  # Sync transports
            return self.execute_sync(
                request,
                serialize_variables=serialize_variables,
                parse_result=parse_result,
                get_execution_result=get_execution_result,
                **kwargs,
            )

    @overload
    def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> List[Dict[str, Any]]: ...  # pragma: no cover

    @overload
    def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> List[ExecutionResult]: ...  # pragma: no cover

    @overload
    def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]: ...  # pragma: no cover

    def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]:
        """Execute multiple GraphQL requests in a batch against the remote server using
        the transport provided during init.

        This function **WILL BLOCK** until the result is received from the server.

        Either the transport is sync and we execute the query synchronously directly
        OR the transport is async and we execute the query in the asyncio loop
        (blocking here until answer).

        This method will:

         - connect using the transport to get a session
         - execute the GraphQL requests on the transport session
         - close the session and close the connection to the server

         If you want to perform multiple executions, it is better to use
         the context manager to keep a session active.

         The extra arguments passed in the method will be passed to the transport
         execute method.
        """

        if isinstance(self.transport, AsyncTransport):
            loop = self._get_event_loop()

            assert not loop.is_running(), (
                "Cannot run client.execute_batch(query) if an asyncio loop is running."
                " Use 'await client.execute_batch(query)' instead."
            )

            data = loop.run_until_complete(
                self.execute_batch_async(
                    requests,
                    serialize_variables=serialize_variables,
                    parse_result=parse_result,
                    get_execution_result=get_execution_result,
                    **kwargs,
                )
            )

            return data

        else:  # Sync transports
            return self.execute_batch_sync(
                requests,
                serialize_variables=serialize_variables,
                parse_result=parse_result,
                get_execution_result=get_execution_result,
                **kwargs,
            )

    @overload
    def subscribe_async(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> AsyncGenerator[Dict[str, Any], None]: ...  # pragma: no cover

    @overload
    def subscribe_async(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> AsyncGenerator[ExecutionResult, None]: ...  # pragma: no cover

    @overload
    def subscribe_async(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[
        AsyncGenerator[Dict[str, Any], None], AsyncGenerator[ExecutionResult, None]
    ]: ...  # pragma: no cover

    async def subscribe_async(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[
        AsyncGenerator[Dict[str, Any], None], AsyncGenerator[ExecutionResult, None]
    ]:
        """:meta private:"""
        async with self as session:
            generator = session.subscribe(
                request,
                serialize_variables=serialize_variables,
                parse_result=parse_result,
                get_execution_result=get_execution_result,
                **kwargs,
            )

            async for result in generator:
                yield result

    @overload
    def subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> Generator[Dict[str, Any], None, None]: ...  # pragma: no cover

    @overload
    def subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> Generator[ExecutionResult, None, None]: ...  # pragma: no cover

    @overload
    def subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[
        Generator[Dict[str, Any], None, None], Generator[ExecutionResult, None, None]
    ]: ...  # pragma: no cover

    def subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[
        Generator[Dict[str, Any], None, None], Generator[ExecutionResult, None, None]
    ]:
        """Execute a GraphQL subscription with a python generator.

        We need an async transport for this functionality.
        """

        loop = self._get_event_loop()

        assert not loop.is_running(), (
            "Cannot run client.subscribe(query) if an asyncio loop is running."
            " Use 'await client.subscribe_async(query)' instead."
        )

        async_generator: Union[
            AsyncGenerator[Dict[str, Any], None], AsyncGenerator[ExecutionResult, None]
        ] = self.subscribe_async(
            request,
            serialize_variables=serialize_variables,
            parse_result=parse_result,
            get_execution_result=get_execution_result,
            **kwargs,
        )

        try:
            while True:
                # Note: we need to create a task here in order to be able to close
                # the async generator properly on python 3.8
                # See https://bugs.python.org/issue38559
                generator_task = asyncio.ensure_future(
                    async_generator.__anext__(), loop=loop
                )
                result: Union[
                    Dict[str, Any], ExecutionResult
                ] = loop.run_until_complete(
                    generator_task
                )  # type: ignore
                yield result

        except StopAsyncIteration:
            pass

        except (KeyboardInterrupt, Exception, GeneratorExit):
            # Graceful shutdown
            asyncio.ensure_future(async_generator.aclose(), loop=loop)

            generator_task.cancel()

            loop.run_until_complete(loop.shutdown_asyncgens())

            # Then reraise the exception
            raise

    async def connect_async(self, reconnecting=False, **kwargs):
        r"""Connect asynchronously with the underlying async transport to
        produce a session.

        That session will be a permanent auto-reconnecting session
        if :code:`reconnecting=True`.

        If you call this method, you should call the
        :meth:`close_async <gql.client.Client.close_async>` method
        for cleanup.

        :param reconnecting: if True, create a permanent reconnecting session
        :param \**kwargs: additional arguments for the
            :meth:`ReconnectingAsyncClientSession init method
            <gql.client.ReconnectingAsyncClientSession.__init__>`.
        """

        assert isinstance(
            self.transport, AsyncTransport
        ), "Only a transport of type AsyncTransport can be used asynchronously"

        self.session: Union[AsyncClientSession, SyncClientSession]

        if reconnecting:
            self.session = ReconnectingAsyncClientSession(client=self, **kwargs)
        else:
            self.session = AsyncClientSession(client=self)

        await self.session.connect()

        # Get schema from transport if needed
        try:
            if self.fetch_schema_from_transport and not self.schema:
                await self.session.fetch_schema()
        except Exception:
            # we don't know what type of exception is thrown here because it
            # depends on the underlying transport; we just make sure that the
            # transport is closed and re-raise the exception
            await self.session.close()
            raise

        return self.session

    async def close_async(self):
        """Close the async transport and stop the optional reconnecting task."""

        await self.session.close()

    async def __aenter__(self):
        return await self.connect_async()

    async def __aexit__(self, exc_type, exc, tb):
        await self.close_async()

    def connect_sync(self):
        r"""Connect synchronously with the underlying sync transport to
        produce a session.

        If you call this method, you should call the
        :meth:`close_sync <gql.client.Client.close_sync>` method
        for cleanup.
        """

        assert not isinstance(self.transport, AsyncTransport), (
            "Only a sync transport can be used."
            " Use 'async with Client(...) as session:' instead"
        )

        if not hasattr(self, "session"):
            self.session = SyncClientSession(client=self)

        assert isinstance(self.session, SyncClientSession)

        self.session.connect()

        # Get schema from transport if needed
        try:
            if self.fetch_schema_from_transport and not self.schema:
                self.session.fetch_schema()
        except Exception:
            # we don't know what type of exception is thrown here because it
            # depends on the underlying transport; we just make sure that the
            # transport is closed and re-raise the exception
            self.session.close()
            raise

        return self.session

    def close_sync(self):
        """Close the sync session and the sync transport.

        If batching is enabled, this will block until the remaining queries in the
        batching queue have been processed.
        """
        assert isinstance(self.session, SyncClientSession)

        self.session.close()

    def __enter__(self):
        return self.connect_sync()

    def __exit__(self, *args):
        self.close_sync()


class SyncClientSession:
    """An instance of this class is created when using :code:`with` on the client.

    It contains the sync method execute to send queries
    on a sync transport using the same session.
    """

    def __init__(self, client: Client):
        """:param client: the :class:`client <gql.client.Client>` used"""
        self.client = client

    def _execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        **kwargs: Any,
    ) -> ExecutionResult:
        """Execute the provided request synchronously using
        the sync transport, returning an ExecutionResult object.

        :param request: GraphQL request as a
                        :class:`GraphQLRequest <gql.GraphQLRequest>` object.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.

        The extra arguments are passed to the transport execute method."""

        # Still supporting for now old method of providing
        # variable_values and operation_name
        request = support_deprecated_request(request, kwargs)

        # Validate document
        if self.client.schema:
            self.client.validate(request)

            # Parse variable values for custom scalars if requested
            if request.variable_values is not None:
                if serialize_variables or (
                    serialize_variables is None and self.client.serialize_variables
                ):
                    request = request.serialize_variable_values(self.client.schema)

        if self.client.batching_enabled:
            future_result = self._execute_future(request)
            result = future_result.result()

        else:
            result = self.transport.execute(
                request,
                **kwargs,
            )

        # Unserialize the result if requested
        if self.client.schema:
            if parse_result or (parse_result is None and self.client.parse_results):
                result.data = parse_result_fn(
                    self.client.schema,
                    request.document,
                    result.data,
                    operation_name=request.operation_name,
                )

        return result

    @overload
    def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> Dict[str, Any]: ...  # pragma: no cover

    @overload
    def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> ExecutionResult: ...  # pragma: no cover

    @overload
    def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]: ...  # pragma: no cover

    def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]:
        """Execute the provided request synchronously using
        the sync transport.

        Raises a TransportQueryError if an error has been returned in
            the ExecutionResult.

        :param request: GraphQL query as :class:`GraphQLRequest <gql.GraphQLRequest>`.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.
        :param get_execution_result: return the full ExecutionResult instance instead of
            only the "data" field. Necessary if you want to get the "extensions" field.

        The extra arguments are passed to the transport execute method."""

        # Validate and execute on the transport
        result = self._execute(
            request,
            serialize_variables=serialize_variables,
            parse_result=parse_result,
            **kwargs,
        )

        # Raise an error if an error is returned in the ExecutionResult object
        if result.errors:
            raise TransportQueryError(
                str_first_element(result.errors),
                errors=result.errors,
                data=result.data,
                extensions=result.extensions,
            )

        assert (
            result.data is not None
        ), "Transport returned an ExecutionResult without data or errors"

        if get_execution_result:
            return result

        return result.data

    def _execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        validate_document: Optional[bool] = True,
        **kwargs: Any,
    ) -> List[ExecutionResult]:
        """Execute multiple GraphQL requests in a batch, using
        the sync transport, returning a list of ExecutionResult objects.

        :param requests: List of requests that will be executed.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.
        :param validate_document: Whether we still need to validate the document.

        The extra arguments are passed to the transport execute method."""

        # Validate document
        if self.client.schema:

            if validate_document:
                for req in requests:
                    self.client.validate(req)

            # Parse variable values for custom scalars if requested
            if serialize_variables or (
                serialize_variables is None and self.client.serialize_variables
            ):
                requests = [
                    (
                        req.serialize_variable_values(self.client.schema)
                        if req.variable_values is not None
                        else req
                    )
                    for req in requests
                ]

        results = self.transport.execute_batch(requests, **kwargs)

        # Unserialize the result if requested
        if self.client.schema:
            if parse_result or (parse_result is None and self.client.parse_results):
                for result in results:
                    result.data = parse_result_fn(
                        self.client.schema,
                        req.document,
                        result.data,
                        operation_name=req.operation_name,
                    )

        return results

    @overload
    def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> List[Dict[str, Any]]: ...  # pragma: no cover

    @overload
    def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> List[ExecutionResult]: ...  # pragma: no cover

    @overload
    def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]: ...  # pragma: no cover

    def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]:
        """Execute multiple GraphQL requests in a batch, using
        the sync transport. This method sends the requests to the server all at once.

        Raises a TransportQueryError if an error has been returned in any
          ExecutionResult.

        :param requests: List of requests that will be executed.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.
        :param get_execution_result: return the full ExecutionResult instance instead of
            only the "data" field. Necessary if you want to get the "extensions" field.

        The extra arguments are passed to the transport execute method."""

        # Validate and execute on the transport
        results = self._execute_batch(
            requests,
            serialize_variables=serialize_variables,
            parse_result=parse_result,
            **kwargs,
        )

        for result in results:
            # Raise an error if an error is returned in the ExecutionResult object
            if result.errors:
                raise TransportQueryError(
                    str_first_element(result.errors),
                    errors=result.errors,
                    data=result.data,
                    extensions=result.extensions,
                )

            assert (
                result.data is not None
            ), "Transport returned an ExecutionResult without data or errors"

        if get_execution_result:
            return results

        return cast(List[Dict[str, Any]], [result.data for result in results])

    def _batch_loop(self) -> None:
        """main loop of the thread used to wait for requests
        to execute them in a batch"""

        stop_loop = False

        while not stop_loop:

            # First wait for a first request in from the batch queue
            requests_and_futures: List[Tuple[GraphQLRequest, Future]] = []
            request_and_future: Tuple[GraphQLRequest, Future] = self.batch_queue.get()
            if request_and_future is None:
                break
            requests_and_futures.append(request_and_future)

            # Then wait the requested batch interval except if we already
            # have the maximum number of requests in the queue
            if self.batch_queue.qsize() < self.client.batch_max - 1:
                time.sleep(self.client.batch_interval)

            # Then get the requests which had been made during that wait interval
            for _ in range(self.client.batch_max - 1):
                if self.batch_queue.empty():
                    break
                request_and_future = self.batch_queue.get()
                if request_and_future is None:
                    stop_loop = True
                    break
                requests_and_futures.append(request_and_future)

            requests = [request for request, _ in requests_and_futures]
            futures = [future for _, future in requests_and_futures]

            # Manually execute the requests in a batch
            try:
                results: List[ExecutionResult] = self._execute_batch(
                    requests,
                    serialize_variables=False,  # already done
                    parse_result=False,
                    validate_document=False,
                )
            except Exception as exc:
                for future in futures:
                    future.set_exception(exc)
                continue

            # Fill in the future results
            for result, future in zip(results, futures):
                future.set_result(result)

        # Indicate that the Thread has stopped
        self._batch_thread_stopped_event.set()

    def _execute_future(
        self,
        request: GraphQLRequest,
    ) -> Future:
        """If batching is enabled, this method will put a request in the batching queue
        instead of executing it directly so that the requests could be put in a batch.
        """

        assert hasattr(self, "batch_queue"), "Batching is not enabled"
        assert not self._batch_thread_stop_requested, "Batching thread has been stopped"

        future: Future = Future()
        self.batch_queue.put((request, future))

        return future

    def connect(self):
        """Connect the transport and initialize the batch threading loop if batching
        is enabled."""

        if self.client.batching_enabled:
            self.batch_queue: Queue = Queue()
            self._batch_thread_stop_requested = False
            self._batch_thread_stopped_event = Event()
            self._batch_thread = Thread(target=self._batch_loop, daemon=True)
            self._batch_thread.start()

        self.transport.connect()

    def close(self):
        """Close the transport and cleanup the batching thread if batching is enabled.

        Will wait until all the remaining requests in the batch processing queue
        have been executed.
        """
        if hasattr(self, "_batch_thread_stopped_event"):
            # Send a None in the queue to indicate that the batching Thread must stop
            # after having processed the remaining requests in the queue
            self._batch_thread_stop_requested = True
            self.batch_queue.put(None)

            # Wait for the Thread to stop
            self._batch_thread_stopped_event.wait()

        self.transport.close()

    def fetch_schema(self) -> None:
        """Fetch the GraphQL schema explicitly using introspection.

        Don't use this function and instead set the fetch_schema_from_transport
        attribute to True"""
        introspection_query = get_introspection_query_ast(
            **self.client.introspection_args
        )
        execution_result = self.transport.execute(GraphQLRequest(introspection_query))

        self.client._build_schema_from_introspection(execution_result)

    @property
    def transport(self):
        return self.client.transport


class AsyncClientSession:
    """An instance of this class is created when using :code:`async with` on a
    :class:`client <gql.client.Client>`.

    It contains the async methods (execute, subscribe) to send queries
    on an async transport using the same session.
    """

    def __init__(self, client: Client):
        """:param client: the :class:`client <gql.client.Client>` used"""
        self.client = client

    async def _subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        **kwargs: Any,
    ) -> AsyncGenerator[ExecutionResult, None]:
        """Coroutine to subscribe asynchronously to the provided request
        asynchronously using the async transport,
        returning an async generator producing ExecutionResult objects.

        * Validate the query with the schema if provided.
        * Serialize the variable_values if requested.

        :param request: GraphQL request as a
                        :class:`GraphQLRequest <gql.GraphQLRequest>` object.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.

        The extra arguments are passed to the transport subscribe method."""

        # Still supporting for now old method of providing
        # variable_values and operation_name
        request = support_deprecated_request(request, kwargs)

        # Validate document
        if self.client.schema:
            self.client.validate(request)

            # Parse variable values for custom scalars if requested
            if request.variable_values is not None:
                if serialize_variables or (
                    serialize_variables is None and self.client.serialize_variables
                ):
                    request = request.serialize_variable_values(self.client.schema)

        # Subscribe to the transport
        inner_generator: AsyncGenerator[ExecutionResult, None] = (
            self.transport.subscribe(
                request,
                **kwargs,
            )
        )

        # Keep a reference to the inner generator
        # This is only used for the tests to simulate a KeyboardInterrupt event
        self._generator = inner_generator

        try:
            async for result in inner_generator:
                if self.client.schema:
                    if parse_result or (
                        parse_result is None and self.client.parse_results
                    ):
                        result.data = parse_result_fn(
                            self.client.schema,
                            request.document,
                            result.data,
                            operation_name=request.operation_name,
                        )

                yield result

        finally:
            await inner_generator.aclose()

    @overload
    def subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> AsyncGenerator[Dict[str, Any], None]: ...  # pragma: no cover

    @overload
    def subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> AsyncGenerator[ExecutionResult, None]: ...  # pragma: no cover

    @overload
    def subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[
        AsyncGenerator[Dict[str, Any], None], AsyncGenerator[ExecutionResult, None]
    ]: ...  # pragma: no cover

    async def subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[
        AsyncGenerator[Dict[str, Any], None], AsyncGenerator[ExecutionResult, None]
    ]:
        """Coroutine to subscribe asynchronously to the provided request
        asynchronously using the async transport.

        Raises a TransportQueryError if an error has been returned in
            the ExecutionResult.

        :param request: GraphQL query as :class:`GraphQLRequest <gql.GraphQLRequest>`.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.
        :param get_execution_result: yield the full ExecutionResult instance instead of
            only the "data" field. Necessary if you want to get the "extensions" field.

        The extra arguments are passed to the transport subscribe method."""

        inner_generator: AsyncGenerator[ExecutionResult, None] = self._subscribe(
            request,
            serialize_variables=serialize_variables,
            parse_result=parse_result,
            **kwargs,
        )

        try:
            # Validate and subscribe on the transport
            async for result in inner_generator:
                # Raise an error if an error is returned in the ExecutionResult object
                if result.errors:
                    raise TransportQueryError(
                        str_first_element(result.errors),
                        errors=result.errors,
                        data=result.data,
                        extensions=result.extensions,
                    )

                elif result.data is not None:
                    if get_execution_result:
                        yield result
                    else:
                        yield result.data
        finally:
            await inner_generator.aclose()

    async def _execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        **kwargs: Any,
    ) -> ExecutionResult:
        """Coroutine to execute the provided request asynchronously using
        the async transport, returning an ExecutionResult object.

        * Validate the query with the schema if provided.
        * Serialize the variable_values if requested.

        :param request: GraphQL request as a
                        :class:`GraphQLRequest <gql.GraphQLRequest>` object.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.

        The extra arguments are passed to the transport execute method."""

        # Still supporting for now old method of providing
        # variable_values and operation_name
        request = support_deprecated_request(request, kwargs)

        # Validate document
        if self.client.schema:
            self.client.validate(request)

            # Parse variable values for custom scalars if requested
            if request.variable_values is not None:
                if serialize_variables or (
                    serialize_variables is None and self.client.serialize_variables
                ):
                    request = request.serialize_variable_values(self.client.schema)

        # Check if batching is enabled
        if self.client.batching_enabled:
            future_result = await self._execute_future(request)
            result = await future_result
        else:
            # Execute the query with the transport with a timeout
            with fail_after(self.client.execute_timeout):
                result = await self.transport.execute(
                    request,
                    **kwargs,
                )

        # Unserialize the result if requested
        if self.client.schema:
            if parse_result or (parse_result is None and self.client.parse_results):
                result.data = parse_result_fn(
                    self.client.schema,
                    request.document,
                    result.data,
                    operation_name=request.operation_name,
                )

        return result

    @overload
    async def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> Dict[str, Any]: ...  # pragma: no cover

    @overload
    async def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> ExecutionResult: ...  # pragma: no cover

    @overload
    async def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = ...,
        parse_result: Optional[bool] = ...,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]: ...  # pragma: no cover

    async def execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], ExecutionResult]:
        """Coroutine to execute the provided request asynchronously using
        the async transport.

        Raises a TransportQueryError if an error has been returned in
            the ExecutionResult.

        :param request: GraphQL query as :class:`GraphQLRequest <gql.GraphQLRequest>`.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.
        :param get_execution_result: return the full ExecutionResult instance instead of
            only the "data" field. Necessary if you want to get the "extensions" field.

        The extra arguments are passed to the transport execute method."""

        # Validate and execute on the transport
        result = await self._execute(
            request,
            serialize_variables=serialize_variables,
            parse_result=parse_result,
            **kwargs,
        )

        # Raise an error if an error is returned in the ExecutionResult object
        if result.errors:
            raise TransportQueryError(
                str_first_element(result.errors),
                errors=result.errors,
                data=result.data,
                extensions=result.extensions,
            )

        assert (
            result.data is not None
        ), "Transport returned an ExecutionResult without data or errors"

        if get_execution_result:
            return result

        return result.data

    async def _execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        validate_document: Optional[bool] = True,
        **kwargs: Any,
    ) -> List[ExecutionResult]:
        """Execute multiple GraphQL requests in a batch, using
        the async transport, returning a list of ExecutionResult objects.

        :param requests: List of requests that will be executed.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.
        :param validate_document: Whether we still need to validate the document.

        The extra arguments are passed to the transport execute_batch method."""

        # Validate document
        if self.client.schema:

            if validate_document:
                for req in requests:
                    self.client.validate(req)

            # Parse variable values for custom scalars if requested
            if serialize_variables or (
                serialize_variables is None and self.client.serialize_variables
            ):
                requests = [
                    (
                        req.serialize_variable_values(self.client.schema)
                        if req.variable_values is not None
                        else req
                    )
                    for req in requests
                ]

        results = await self.transport.execute_batch(requests, **kwargs)

        # Unserialize the result if requested
        if self.client.schema:
            if parse_result or (parse_result is None and self.client.parse_results):
                for result in results:
                    result.data = parse_result_fn(
                        self.client.schema,
                        req.document,
                        result.data,
                        operation_name=req.operation_name,
                    )

        return results

    @overload
    async def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[False] = ...,
        **kwargs: Any,
    ) -> List[Dict[str, Any]]: ...  # pragma: no cover

    @overload
    async def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: Literal[True],
        **kwargs: Any,
    ) -> List[ExecutionResult]: ...  # pragma: no cover

    @overload
    async def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]: ...  # pragma: no cover

    async def execute_batch(
        self,
        requests: List[GraphQLRequest],
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        get_execution_result: bool = False,
        **kwargs: Any,
    ) -> Union[List[Dict[str, Any]], List[ExecutionResult]]:
        """Execute multiple GraphQL requests in a batch, using
        the async transport. This method sends the requests to the server all at once.

        Raises a TransportQueryError if an error has been returned in any
          ExecutionResult.

        :param requests: List of requests that will be executed.
        :param serialize_variables: whether the variable values should be
            serialized. Used for custom scalars and/or enums.
            By default use the serialize_variables argument of the client.
        :param parse_result: Whether gql will deserialize the result.
            By default use the parse_results argument of the client.
        :param get_execution_result: return the full ExecutionResult instance instead of
            only the "data" field. Necessary if you want to get the "extensions" field.

        The extra arguments are passed to the transport execute method."""

        # Validate and execute on the transport
        results = await self._execute_batch(
            requests,
            serialize_variables=serialize_variables,
            parse_result=parse_result,
            **kwargs,
        )

        for result in results:
            # Raise an error if an error is returned in the ExecutionResult object
            if result.errors:
                raise TransportQueryError(
                    str_first_element(result.errors),
                    errors=result.errors,
                    data=result.data,
                    extensions=result.extensions,
                )

            assert (
                result.data is not None
            ), "Transport returned an ExecutionResult without data or errors"

        if get_execution_result:
            return results

        return cast(List[Dict[str, Any]], [result.data for result in results])

    async def _batch_loop(self) -> None:
        """Main loop of the task used to wait for requests
        to execute them in a batch"""

        stop_loop = False

        while not stop_loop:
            # First wait for a first request in from the batch queue
            requests_and_futures: List[Tuple[GraphQLRequest, asyncio.Future]] = []

            # Wait for the first request
            request_and_future: Optional[Tuple[GraphQLRequest, asyncio.Future]] = (
                await self.batch_queue.get()
            )

            if request_and_future is None:
                # None is our sentinel value to stop the loop
                break

            requests_and_futures.append(request_and_future)

            # Then wait the requested batch interval except if we already
            # have the maximum number of requests in the queue
            if self.batch_queue.qsize() < self.client.batch_max - 1:
                # Wait for the batch interval
                await asyncio.sleep(self.client.batch_interval)

            # Then get the requests which had been made during that wait interval
            for _ in range(self.client.batch_max - 1):
                try:
                    # Use get_nowait since we don't want to wait here
                    request_and_future = self.batch_queue.get_nowait()

                    if request_and_future is None:
                        # Sentinel value - stop after processing current batch
                        stop_loop = True
                        break

                    requests_and_futures.append(request_and_future)

                except asyncio.QueueEmpty:
                    # No more requests in queue, that's fine
                    break

            # Extract requests and futures
            requests = [request for request, _ in requests_and_futures]
            futures = [future for _, future in requests_and_futures]

            # Execute the batch
            try:
                results: List[ExecutionResult] = await self._execute_batch(
                    requests,
                    serialize_variables=False,  # already done
                    parse_result=False,  # will be done later
                    validate_document=False,  # already validated
                )

                # Set the result for each future
                for result, future in zip(results, futures):
                    if not future.cancelled():
                        future.set_result(result)

            except Exception as exc:
                # If batch execution fails, propagate the error to all futures
                for future in futures:
                    if not future.cancelled():
                        future.set_exception(exc)

        # Signal that the task has stopped
        self._batch_task_stopped_event.set()

    async def _execute_future(
        self,
        request: GraphQLRequest,
    ) -> asyncio.Future:
        """If batching is enabled, this method will put a request in the batching queue
        instead of executing it directly so that the requests could be put in a batch.
        """

        assert hasattr(self, "batch_queue"), "Batching is not enabled"
        assert not self._batch_task_stop_requested, "Batching task has been stopped"

        future: asyncio.Future = asyncio.Future()
        await self.batch_queue.put((request, future))

        return future

    async def _batch_init(self):
        """Initialize the batch task loop if batching is enabled."""
        if self.client.batching_enabled:
            self.batch_queue: asyncio.Queue = asyncio.Queue()
            self._batch_task_stop_requested = False
            self._batch_task_stopped_event = asyncio.Event()
            self._batch_task = asyncio.create_task(self._batch_loop())

    async def _batch_cleanup(self):
        """Cleanup the batching task if batching is enabled."""
        if hasattr(self, "_batch_task_stopped_event"):
            # Send a None in the queue to indicate that the batching task must stop
            # after having processed the remaining requests in the queue
            self._batch_task_stop_requested = True
            await self.batch_queue.put(None)

            # Wait for the task to process remaining requests and stop
            await self._batch_task_stopped_event.wait()

    async def connect(self):
        """Connect the transport and initialize the batch task loop if batching
        is enabled."""

        await self._batch_init()

        try:
            await self.transport.connect()
        except Exception as e:
            await self.transport.close()
            raise e

    async def close(self):
        """Close the transport and cleanup the batching task if batching is enabled.

        Will wait until all the remaining requests in the batch processing queue
        have been executed.
        """
        await self._batch_cleanup()

        await self.transport.close()

    async def fetch_schema(self) -> None:
        """Fetch the GraphQL schema explicitly using introspection.

        Don't use this function and instead set the fetch_schema_from_transport
        attribute to True"""
        introspection_query = get_introspection_query_ast(
            **self.client.introspection_args
        )
        execution_result = await self.transport.execute(
            GraphQLRequest(introspection_query)
        )

        self.client._build_schema_from_introspection(execution_result)

    @property
    def transport(self):
        return self.client.transport


_CallableT = TypeVar("_CallableT", bound=Callable[..., Any])
_Decorator = Callable[[_CallableT], _CallableT]


class ReconnectingAsyncClientSession(AsyncClientSession):
    """An instance of this class is created when using the
    :meth:`connect_async <gql.client.Client.connect_async>` method of the
    :class:`Client <gql.client.Client>` class with :code:`reconnecting=True`.

    It is used to provide a single session which will reconnect automatically if
    the connection fails.
    """

    def __init__(
        self,
        client: Client,
        *,
        retry_connect: Union[bool, _Decorator] = True,
        retry_execute: Union[bool, _Decorator] = True,
    ):
        """
        :param client: the :class:`client <gql.client.Client>` used.
        :param retry_connect: Either a Boolean to activate/deactivate the retries
            for the connection to the transport OR a backoff decorator to
            provide specific retries parameters for the connections.
        :param retry_execute: Either a Boolean to activate/deactivate the retries
            for the execute method OR a backoff decorator to
            provide specific retries parameters for this method.
        """
        self.client = client
        self._connect_task = None

        self._reconnect_request_event = asyncio.Event()
        self._connected_event = asyncio.Event()

        if retry_connect is True:
            # By default, retry again and again, with maximum 60 seconds
            # between retries
            self.retry_connect = backoff.on_exception(
                backoff.expo,
                Exception,
                max_value=60,
            )
        elif retry_connect is False:
            self.retry_connect = lambda e: e
        else:
            assert callable(retry_connect)
            self.retry_connect = retry_connect

        if retry_execute is True:
            # By default, retry 5 times, except if we receive a TransportQueryError
            self.retry_execute = backoff.on_exception(
                backoff.expo,
                Exception,
                max_tries=5,
                giveup=lambda e: isinstance(e, TransportQueryError),
            )
        elif retry_execute is False:
            self.retry_execute = lambda e: e
        else:
            assert callable(retry_execute)
            self.retry_execute = retry_execute

        # Creating the _execute_with_retries and _connect_with_retries  methods
        # using the provided backoff decorators
        self._execute_with_retries = self.retry_execute(self._execute_once)
        self._connect_with_retries = self.retry_connect(self.transport.connect)

    async def _connection_loop(self):
        """Coroutine used for the connection task.

        - try to connect to the transport with retries
        - send a connected event when the connection has been made
        - then wait for a reconnect request to try to connect again
        """

        while True:
            # Connect to the transport with the retry decorator
            # By default it should keep retrying until it connect
            await self._connect_with_retries()

            # Once connected, set the connected event
            self._connected_event.set()
            self._connected_event.clear()

            # Then wait for the reconnect event
            self._reconnect_request_event.clear()
            await self._reconnect_request_event.wait()
            await self.transport.close()

    async def start_connecting_task(self):
        """Start the task responsible to restart the connection
        of the transport when requested by an event.
        """
        if self._connect_task:
            log.warning("connect task already started!")
        else:
            self._connect_task = asyncio.ensure_future(self._connection_loop())

            await self._connected_event.wait()

    async def stop_connecting_task(self):
        """Stop the connecting task."""
        if self._connect_task is not None:
            self._connect_task.cancel()
            self._connect_task = None

    async def connect(self):
        """Start the connect task and initialize the batch task loop if batching
        is enabled."""

        await self._batch_init()

        await self.start_connecting_task()

    async def close(self):
        """Stop the connect task and cleanup the batching task
        if batching is enabled."""
        await self._batch_cleanup()

        await self.stop_connecting_task()

        await self.transport.close()

    async def _execute_once(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        **kwargs: Any,
    ) -> ExecutionResult:
        """Same Coroutine as parent method _execute but requesting a
        reconnection if we receive a TransportConnectionFailed exception.
        """

        try:
            answer = await super()._execute(
                request,
                serialize_variables=serialize_variables,
                parse_result=parse_result,
                **kwargs,
            )
        except TransportConnectionFailed:
            self._reconnect_request_event.set()
            raise

        return answer

    async def _execute(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        **kwargs: Any,
    ) -> ExecutionResult:
        """Same Coroutine as parent, but with optional retries
        and requesting a reconnection if we receive a
        TransportConnectionFailed exception.
        """

        return await self._execute_with_retries(
            request,
            serialize_variables=serialize_variables,
            parse_result=parse_result,
            **kwargs,
        )

    async def _subscribe(
        self,
        request: GraphQLRequest,
        *,
        serialize_variables: Optional[bool] = None,
        parse_result: Optional[bool] = None,
        **kwargs: Any,
    ) -> AsyncGenerator[ExecutionResult, None]:
        """Same Async generator as parent method _subscribe but requesting a
        reconnection if we receive a TransportConnectionFailed exception.
        """

        inner_generator: AsyncGenerator[ExecutionResult, None] = super()._subscribe(
            request,
            serialize_variables=serialize_variables,
            parse_result=parse_result,
            **kwargs,
        )

        try:
            async for result in inner_generator:
                yield result

        except TransportConnectionFailed:
            self._reconnect_request_event.set()
            raise

        finally:
            await inner_generator.aclose()



import asyncio
import os
import sys
from urllib.parse import urlparse

from gql import Client, gql
from gql.transport.appsync_auth import AppSyncApiKeyAuthentication
from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

# Uncomment the following lines to enable debug output
# import logging
# logging.basicConfig(level=logging.DEBUG)


async def main():

    # Should look like:
    # https://XXXXXXXXXXXXXXXXXXXXXXXXXX.appsync-api.REGION.amazonaws.com/graphql
    url = os.environ.get("AWS_GRAPHQL_API_ENDPOINT")
    api_key = os.environ.get("AWS_GRAPHQL_API_KEY")

    if url is None or api_key is None:
        print("Missing environment variables")
        sys.exit()

    # Extract host from url
    host = str(urlparse(url).netloc)

    print(f"Host: {host}")

    auth = AppSyncApiKeyAuthentication(host=host, api_key=api_key)

    transport = AppSyncWebsocketsTransport(url=url, auth=auth)

    async with Client(transport=transport) as session:

        subscription = gql(
            """
subscription onCreateMessage {
  onCreateMessage {
    message
  }
}
"""
        )

        print("Waiting for messages...")

        async for result in session.subscribe(subscription):
            print(result)


asyncio.run(main())



from itertools import repeat

from graphql import DocumentNode, GraphQLSchema

from gql.dsl import DSLFragment, DSLMetaField, DSLQuery, DSLSchema, dsl_gql


def get_introspection_query_ast(
    descriptions: bool = True,
    specified_by_url: bool = False,
    directive_is_repeatable: bool = False,
    schema_description: bool = False,
    input_value_deprecation: bool = True,
    type_recursion_level: int = 7,
) -> DocumentNode:
    """Get a query for introspection as a document using the DSL module.

    Equivalent to the get_introspection_query function from graphql-core
    but using the DSL module and allowing to select the recursion level.

    Optionally, you can exclude descriptions, include specification URLs,
    include repeatability of directives, and specify whether to include
    the schema description as well.
    """

    ds = DSLSchema(GraphQLSchema())

    fragment_FullType = DSLFragment("FullType").on(ds.__Type)
    fragment_InputValue = DSLFragment("InputValue").on(ds.__InputValue)
    fragment_TypeRef = DSLFragment("TypeRef").on(ds.__Type)

    schema = DSLMetaField("__schema")

    if descriptions and schema_description:
        schema.select(ds.__Schema.description)

    schema.select(
        ds.__Schema.queryType.select(ds.__Type.name),
        ds.__Schema.mutationType.select(ds.__Type.name),
        ds.__Schema.subscriptionType.select(ds.__Type.name),
    )

    schema.select(ds.__Schema.types.select(fragment_FullType))

    directives = ds.__Schema.directives.select(ds.__Directive.name)

    deprecated_expand = {}

    if input_value_deprecation:
        deprecated_expand = {
            "includeDeprecated": True,
        }

    if descriptions:
        directives.select(ds.__Directive.description)
    if directive_is_repeatable:
        directives.select(ds.__Directive.isRepeatable)
    directives.select(
        ds.__Directive.locations,
        ds.__Directive.args(**deprecated_expand).select(fragment_InputValue),
    )

    schema.select(directives)

    fragment_FullType.select(
        ds.__Type.kind,
        ds.__Type.name,
    )
    if descriptions:
        fragment_FullType.select(ds.__Type.description)
    if specified_by_url:
        fragment_FullType.select(ds.__Type.specifiedByURL)

    fields = ds.__Type.fields(includeDeprecated=True).select(ds.__Field.name)

    if descriptions:
        fields.select(ds.__Field.description)

    fields.select(
        ds.__Field.args(**deprecated_expand).select(fragment_InputValue),
        ds.__Field.type.select(fragment_TypeRef),
        ds.__Field.isDeprecated,
        ds.__Field.deprecationReason,
    )

    enum_values = ds.__Type.enumValues(includeDeprecated=True).select(
        ds.__EnumValue.name
    )

    if descriptions:
        enum_values.select(ds.__EnumValue.description)

    enum_values.select(
        ds.__EnumValue.isDeprecated,
        ds.__EnumValue.deprecationReason,
    )

    fragment_FullType.select(
        fields,
        ds.__Type.inputFields(**deprecated_expand).select(fragment_InputValue),
        ds.__Type.interfaces.select(fragment_TypeRef),
        enum_values,
        ds.__Type.possibleTypes.select(fragment_TypeRef),
    )

    fragment_InputValue.select(ds.__InputValue.name)

    if descriptions:
        fragment_InputValue.select(ds.__InputValue.description)

    fragment_InputValue.select(
        ds.__InputValue.type.select(fragment_TypeRef),
        ds.__InputValue.defaultValue,
    )

    if input_value_deprecation:
        fragment_InputValue.select(
            ds.__InputValue.isDeprecated,
            ds.__InputValue.deprecationReason,
        )

    fragment_TypeRef.select(
        ds.__Type.kind,
        ds.__Type.name,
    )

    if type_recursion_level >= 1:
        current_field = ds.__Type.ofType.select(ds.__Type.kind, ds.__Type.name)
        fragment_TypeRef.select(current_field)

        for _ in repeat(None, type_recursion_level - 1):
            new_oftype = ds.__Type.ofType.select(ds.__Type.kind, ds.__Type.name)
            current_field.select(new_oftype)
            current_field = new_oftype

    query = DSLQuery(schema)

    query.name = "IntrospectionQuery"

    dsl_query = dsl_gql(query, fragment_FullType, fragment_InputValue, fragment_TypeRef)

    return dsl_query.document
